import numpy as np
from scipy.special import kv as beselk

from . import SpDust_data_dir
from .util import cgsconst, makelogtab, maketab, DX_over_X
from .Grain import acx, Inertia_largest, grainparams

import os
from .mpiutil import *


k = cgsconst.k
c = cgsconst.c
mp = cgsconst.mp
q = cgsconst.q
pi = np.pi
I = 1j  # complex number equivalent to IDL's "I = complex(0d, 1d)"

class smalletabs:
    smalle_tab = None
    Gamma_tab = None
    Gamma_max = None

class plasma_tabs:
    Ipos_tab = None
    Ineg_tab = None
    rot_tab = None
    e_1tab = None
    rot_min = None
    rot_max = None
    e_1min = None
    e_1max = None

class gp_arrays:
    psi_min = 1e-5
    psi_max = 1e6
    Npsi    = 110

    phi_min = 1e-2
    phi_max = 5e2
    Nphi    = 300

    Omega_min = 1e-10
    Omega_max = 1e5
    NOmega    = 150

    gp_pos, gp_neg, gp_neutral = None, None, None


def compute_small_e():
    
    # Create arrays
    gamma_min = 1e-20
    gamma_max = 10 ** 1.5
    Ngamma = 300
    Gamma_tab = makelogtab(gamma_min, gamma_max, Ngamma)
    smalle_tab = np.zeros(Ngamma)  # Equivalent to dblarr in IDL
    
    # Parameters
    Ny = 20000
    ymed = 1.0
    ymax = 1e20
    
    # Create u and Du arrays
    u_part_1 = maketab(0, ymed, Ny)
    u_part_2 = makelogtab(ymed, ymax, Ny)
    u = np.exp(I * pi / 6) * np.concatenate((u_part_1, u_part_2))
    
    # Differential array Du
    Du_part_1 = ymed / Ny + np.zeros(Ny)
    Du_part_2 = DX_over_X(ymed, ymax, Ny) * u_part_2
    Du = np.exp(I * pi / 6) * np.concatenate((Du_part_1, Du_part_2))
    
    # Function components
    fcos = (u ** 2 - 1) / (u ** 2 + 1) ** 2
    fsin = u / (u ** 2 + 1) ** 2
    
    # Main computation loop
    """
    for ig in range(Ngamma):
        Gamma = Gamma_tab[ig]
        time = Gamma * (u + u ** 3 / 3)
        
        # Integrals
        intcos = 4 * np.sum(np.real(np.exp(I * time) * fcos * Du)) ** 2
        intsin = 16 * np.sum(np.imag(np.exp(I * time) * fsin * Du)) ** 2
        
        # Store results in smalle_tab
        smalle_tab[ig] = intcos + intsin    
    """

    # Define function for parallel computation
    def compute_small_e_parallel(ig):
        Gamma = Gamma_tab[ig]
        time = Gamma * (u + u ** 3 / 3)
        
        # Integrals
        intcos = 4 * np.sum(np.real(np.exp(I * time) * fcos * Du)) ** 2
        intsin = 16 * np.sum(np.imag(np.exp(I * time) * fsin * Du)) ** 2
        
        return intcos + intsin
    
    # Parallel computation
    smalle_tab = np.array(parallel_map(compute_small_e_parallel, list(range(Ngamma))))
    
    # Output the maximum gamma for debugging purposes
    gamma_max = np.max(Gamma_tab)
    
    return Gamma_tab, smalle_tab, gamma_max

# Call `compute_small_e` to get the values for small e_1, Zg < 0
Gamma_tab, smalle_tab, Gamma_max = compute_small_e()
smalletabs.Gamma_tab = Gamma_tab
smalletabs.smalle_tab = smalle_tab
smalletabs.Gamma_max = Gamma_max
if rank0:
    print("I(Zg<0, parabolic) stored")

#@njit
def replace_zeros(Ipos_tab):
    """
    Replace zeros in a 2D array with a small value (1e-30), element by element.
    """
    # Iterate through the array and replace zeros
    for i in range(Ipos_tab.shape[0]):
        for j in range(Ipos_tab.shape[1]):
            if Ipos_tab[i, j] == 0:
                Ipos_tab[i, j] = 1e-30

    return Ipos_tab

def compute_int_plasma():
    file_path = os.path.join(SpDust_data_dir, 'int_plasma.npz')
    if os.path.exists(file_path):
        data = np.load(file_path)
        Ipos_tab = data['Ipos_tab']
        Ineg_tab = data['Ineg_tab']
        rot_tab = data['rot_tab']
        e_1tab = data['e_1tab']
        rot_min = data['rot_min']
        rot_max = data['rot_max']
        e_1min = data['e_1min']
        e_1max = data['e_1max']
        return Ipos_tab, Ineg_tab, rot_tab, e_1tab, rot_min, rot_max, e_1min, e_1max
    
    # Create arrays
    e_1min = 1e-15
    e_1max = 1e4
    Ne_1 = 100
    rot_min = 1e-7
    rot_max = 1.0 / e_1min * Gamma_max
    Nrot = 100

    e_1small = 1e-2
    rot_small = 1e3

    rot_tab = makelogtab(rot_min, rot_max, Nrot)
    e_1tab = makelogtab(e_1min, e_1max, Ne_1)
    Ipos_tab = np.zeros((Nrot, Ne_1))
    Ineg_tab = np.zeros((Nrot, Ne_1))

    # Parameters
    Ny = 10000
    ymax = 1e50
    ymed = 800.0

    y= np.zeros(2*Ny)
    y[:Ny] = maketab(0, ymed, Ny)
    y[Ny:] = makelogtab(ymed, ymax, Ny)

    # Zg > 0 case
    Dz1 = -I * np.concatenate((ymed / Ny + np.zeros(Ny), DX_over_X(ymed, ymax, Ny) * makelogtab(ymed, ymax, Ny)))
    log = 0.5 * np.log(1.0 + 4.0 / y**2) + I * np.arctan(2.0 / y)
    z = 1.0 - I * y

    '''
    for ie in range(Ne_1):
        e_1 = e_1tab[ie]
        Aval = e_1 / (e_1 + 2.0)
        time = 1.0 / np.sqrt(e_1 * (e_1 + 2.0)) * (log + 2.0 * (e_1 + 1.0) * z / (z**2 - 1.0))
        fcos = (z**2 - Aval) / (z**2 + Aval)**2
        fsin = z / (z**2 + Aval)**2
        for ir in range(Nrot):
            rot = rot_tab[ir]
            intcos = 4.0 * Aval * np.sum(np.real(np.exp(I * rot * time) * fcos * Dz1))**2
            intsin = 16.0 * Aval**2 * np.sum(np.imag(np.exp(I * rot * time) * fsin * Dz1))**2
            Ipos_tab[ir, ie] = intcos + intsin
    '''

    # Define function for parallel computation
    #@jit(nopython=True)
    def loop(ie):
        e_1 = e_1tab[ie]
        Aval = e_1 / (e_1 + 2.0)
        time = 1.0 / np.sqrt(e_1 * (e_1 + 2.0)) * (log + 2.0 * (e_1 + 1.0) * z / (z**2 - 1.0))
        fcos = (z**2 - Aval) / (z**2 + Aval)**2
        fsin = z / (z**2 + Aval)**2
        result = np.zeros(Nrot)
        for ir in range(Nrot):
            rot = rot_tab[ir]
            intcos = 4.0 * Aval * np.sum(np.real(np.exp(I * rot * time) * fcos * Dz1))**2
            intsin = 16.0 * Aval**2 * np.sum(np.imag(np.exp(I * rot * time) * fsin * Dz1))**2
            result[ir] = intcos + intsin
        return result
    
    # Parallel computation
    Ipos_tab = np.array(parallel_map(loop, list(range(Ne_1)))).T

    # Handling small e_1 and high rotation for Zg < 0
    inde = np.where(e_1tab < e_1small)[0][-1]
    indrot = np.where(rot_tab < rot_small)[0][-1]

    # Zg < 0 non-small e_1
    Dz2 = np.exp(I * pi / 4.0) * np.concatenate((ymed / Ny + np.zeros(Ny), DX_over_X(ymed, ymax, Ny) * makelogtab(ymed, ymax, Ny)))
    log = 0.5 * np.log(1.0 + 4.0 / y**2 + 2.0 * np.sqrt(2.0) / y) - I * np.arctan(np.sqrt(2.0) / (np.sqrt(2.0) + y))
    z = 1.0 + np.exp(I * pi / 4.0) * y

    # Define function for parallel computation
    #(nopython=True)
    def loop(ie):
        e_1 = e_1tab[ie]
        Aval = e_1 / (e_1 + 2.0)
        time = 1.0 / np.sqrt(e_1 * (e_1 + 2.0)) * (log - 2.0 * (e_1 + 1.0) * z / (z**2 - 1.0))
        fcos = (1.0 - Aval * z**2) / (1.0 + Aval * z**2)**2
        fsin = z / (1.0 + Aval * z**2)**2
        result = np.zeros(indrot)
        for ir in range(indrot):
            rot = rot_tab[ir]
            intcos = 4.0 * Aval * np.sum(np.real(np.exp(I * rot * time) * fcos * Dz2))**2
            intsin = 16.0 * Aval**2 * np.sum(np.imag(np.exp(I * rot * time) * fsin * Dz2))**2
            result[ir] = intcos + intsin
        return result
    
    # Parallel computation
    Ineg_tab[:indrot,:] = np.array(parallel_map(loop, list(range(Ne_1)))).T

    # Define function for parallel computation
    #@jit(nopython=True)
    def loop(ie):
        e_1 = e_1tab[ie]
        result = np.zeros(Nrot)
        for ir in range(indrot, Nrot):
            rot = rot_tab[ir]
            Gamma = rot * e_1
            if Gamma < Gamma_max:
                result[ir] = np.interp(Gamma, Gamma_tab, smalle_tab)
        return result
    
    # Parallel computation
    if Nrot > indrot:
        result = np.array(parallel_map(loop, list(range(inde)))).T
        Ineg_tab[indrot:, :inde] = result[indrot:, :]

    # Some useful things for further interpolation
    rot_min = np.min(rot_tab)
    rot_max = np.max(rot_tab)
    e_1min = np.min(e_1tab)
    e_1max = np.max(e_1tab)

    # Replace zeros with small values for precision
    Ipos_tab = replace_zeros(Ipos_tab)
    Ineg_tab = replace_zeros(Ineg_tab)

    # Save the arrays to file
    if rank0:
        np.savez(file_path, Ipos_tab=Ipos_tab, Ineg_tab=Ineg_tab, rot_tab=rot_tab, e_1tab=e_1tab,
             rot_min=rot_min, rot_max=rot_max, e_1min=e_1min, e_1max=e_1max)

    return Ipos_tab, Ineg_tab, rot_tab, e_1tab, rot_min, rot_max, e_1min, e_1max



Ipos_tab, Ineg_tab, rot_tab, e_1tab, rot_min, rot_max, e_1min, e_1max = compute_int_plasma()

plasma_tabs.Ipos_tab = Ipos_tab
plasma_tabs.Ineg_tab = Ineg_tab
plasma_tabs.rot_tab = rot_tab
plasma_tabs.e_1tab = e_1tab
plasma_tabs.rot_min = rot_min
plasma_tabs.rot_max = rot_max
plasma_tabs.e_1min = e_1min
plasma_tabs.e_1max = e_1max
if rank0:
    print('I(rot, e, Zg <> 0) stored')

barrier()

def int_plasma(rot_new, e_1_new, Zg):
    """
    Returns I(omega * b/v, e-1, Zg non zero) by interpolating arrays computed by `compute_int_plasma`.

    Parameters:
    - rot_new: Array of new rotation values.
    - e_1_new: Array of new e_1 values.
    - Zg: Grain charge (Zg > 0 or Zg < 0).
    - Ipos_tab, Ineg_tab: Precomputed positive and negative I tables.
    - rot_tab, e_1tab: Precomputed rotation and e_1 tables.
    - rot_min, rot_max: Minimum and maximum rotation values.
    - e_1min, e_1max: Minimum and maximum e_1 values.

    Returns:
    - result: Interpolated integrals for the input values.
    """          

    # Ensure rot_new and e_1_new have the same number of elements
    Nelem = np.size(rot_new)
    if Nelem != np.size(e_1_new):
        raise ValueError(f"Error in int_plasma: e_1_new and rot_new should have the same dimensions. "
                         f"Nelem (rot_new) = {Nelem}, N_elements(e_1_new) = {np.size(e_1_new)}")
    
    result = np.zeros(Nelem)

    # Case: e_1_new >= e_1max and rot_new < 100
    ind = np.where((e_1_new >= e_1max) & (rot_new < 100))[0]
    if np.size(ind) > 0:
        rot = rot_new[ind]
        result[ind] = rot**2 * (beselk(0, rot)**2 + beselk(1, rot)**2)

    # Case: rot_new <= rot_min
    ind = np.where(rot_new <= rot_min)[0]
    if np.size(ind) > 0:
        e_1 = e_1_new[ind]
        result[ind] = e_1 * (e_1 + 2) / (e_1 + 1)**2

    # Case: Intermediate values within computed tables' range
    ind = np.where((rot_new < rot_max) & (rot_new > rot_min) & 
                   (e_1_new < e_1max) & (e_1_new > e_1min))[0]
    if np.size(ind) > 0:
        rot = rot_new[ind]
        e_1 = e_1_new[ind]
        
        # Interpolation indices and interpolation factors
        Drot_over_rot = np.log(rot_tab[1] / rot_tab[0])
        irot = np.int64(np.floor(np.log(rot / np.min(rot_tab)) / Drot_over_rot))
        
        De_over_e = np.log(e_1tab[1] / e_1tab[0])
        ie = np.int64(np.floor(np.log(e_1 / np.min(e_1tab)) / De_over_e))
        
        alpha = np.log(rot / rot_tab[irot]) / Drot_over_rot
        beta = np.log(e_1 / e_1tab[ie]) / De_over_e
        
        if Zg > 0:
            result[ind] = np.exp(alpha * (beta * np.log(Ipos_tab[irot + 1, ie + 1]) + (1 - beta) * np.log(Ipos_tab[irot + 1, ie])) +
                                 (1 - alpha) * (beta * np.log(Ipos_tab[irot, ie + 1]) + (1 - beta) * np.log(Ipos_tab[irot, ie])))
        else:
            result[ind] = np.exp(alpha * (beta * np.log(Ineg_tab[irot + 1, ie + 1]) + (1 - beta) * np.log(Ineg_tab[irot + 1, ie])) +
                                 (1 - alpha) * (beta * np.log(Ineg_tab[irot, ie + 1]) + (1 - beta) * np.log(Ineg_tab[irot, ie])))

    return result

def little_gp_charged(psi, Omega):
    u_min = 1e-10
    u_max = 5
    Nu = 250
    c_min = 1e-10
    c_max = 5e10
    Nc = 250

    # Create u_1d and differential over u
    u_1d = makelogtab(u_min, u_max, Nu)
    Du_over_u = DX_over_X(u_min, u_max, Nu)

    # Create arrays for u and c
    u_arr = np.repeat(u_1d, Nc)  # This simulates the column by row expansion
    c_arr = np.zeros(Nc * Nu)
    Dc_over_c = np.zeros(Nc * Nu)

    # Positively charged grains
    if psi > 0:
        for iu in range(Nu):
            u = u_1d[iu]
            if u**2 <= psi:
                c1 = c_min
            else:
                c1 = np.sqrt(1.0 - psi / u**2)
            
            if c1 < c_max:
                c_arr[iu * Nc:(iu + 1) * Nc] = makelogtab(c1, c_max, Nc)
                Dc_over_c[iu * Nc:(iu + 1) * Nc] = DX_over_X(c1, c_max, Nc)
        
        # Calculations for positively charged grains
        stuff = (2.0 / psi * c_arr * u_arr**2)**2
        e_1arr = np.sqrt(1.0 + stuff) - 1.0
        ind = np.where(stuff < 1e-10)
        if np.size(ind[0]) > 0:
            e_1arr[ind] = 0.5 * stuff[ind]

        # Rotation and plasma integral
        rot_arr = Omega * c_arr / u_arr
        Int_arr = int_plasma(rot_arr, e_1arr, 1.0)

        # Return the final result
        return 2.0 * np.sum(u_arr**2 * np.exp(-u_arr**2) * Int_arr * Dc_over_c * Du_over_u)

    # Negatively charged grains
    elif psi < 0:
        for iu in range(Nu):
            u = u_1d[iu]
            c1 = np.sqrt(1.0 - psi / u**2)
            if c1 < c_max:
                c_arr[iu * Nc:(iu + 1) * Nc] = makelogtab(c1, c_max, Nc)
                Dc_over_c[iu * Nc:(iu + 1) * Nc] = DX_over_X(c1, c_max, Nc)

        # Calculations for negatively charged grains
        stuff = (2.0 / psi * c_arr * u_arr**2)**2
        e_1arr = np.sqrt(1.0 + stuff) - 1.0
        ind = np.where(stuff < 1e-10)
        if np.size(ind[0]) > 0:
            e_1arr[ind] = 0.5 * stuff[ind]

        # Rotation and plasma integral
        rot_arr = Omega * c_arr / u_arr
        Int_arr = int_plasma(rot_arr, e_1arr, -1.0)

        # Return the final result
        return 2.0 * Du_over_u * np.sum(u_arr**2 * np.exp(-u_arr**2) * Int_arr * Dc_over_c)

    # If psi == 0, no result
    return None

def little_gp_neutral(phi, Omega):
    u_min = 1e-10
    u_max = 5.0
    Nu = 250
    X_max = 20.0
    NX = 250

    # Logarithmically spaced u values
    u_tab = makelogtab(u_min, u_max, Nu)
    Du_over_u = DX_over_X(u_min, u_max, Nu)
    
    # Array to store the integral for each u
    Int_c = np.zeros(Nu)

    '''
    # Loop over u values
    for iu in range(Nu):
        u = u_tab[iu]
        X_min = Omega / u * np.sqrt(1.0 + phi / u)

        if X_min < X_max:
            # Logarithmically spaced X values
            X_tab = makelogtab(X_min, X_max, NX)
            DX_over_X_val = DX_over_X(X_min, X_max, NX)

            # Compute the integral for this u
            Int_c[iu] = np.sum(X_tab**2 * (beselk(0, X_tab)**2 + beselk(1, X_tab)**2) * DX_over_X_val)
    '''

    # Define function for parallel computation
    def compute_integral(iu):
        u = u_tab[iu]
        X_min = Omega / u * np.sqrt(1.0 + phi / u)
        if X_min < X_max:
            X_tab = makelogtab(X_min, X_max, NX)
            DX_over_X_val = DX_over_X(X_min, X_max, NX)
            return np.sum(X_tab**2 * (beselk(0, X_tab)**2 + beselk(1, X_tab)**2) * DX_over_X_val)
        else:
            return 0.0
        
    # Parallel computation
    Int_c = np.array(parallel_map(compute_integral, list(range(Nu))))

    # Final integral over u
    result = 2.0 * Du_over_u * np.sum(u_tab**2 * np.exp(-u_tab**2) * Int_c )

    return result

def compute_little_gp():
    # If the arrays have been computed and stored, load them
    gp_pos_filename = f"{SpDust_data_dir}/gp_pos_{gp_arrays.Npsi}psi_{gp_arrays.NOmega}Omega.txt"
    gp_neg_filename = f"{SpDust_data_dir}/gp_neg_{gp_arrays.Npsi}psi_{gp_arrays.NOmega}Omega.txt"
    gp_neutral_filename = f"{SpDust_data_dir}/gp_neutral_{gp_arrays.Nphi}phi_{gp_arrays.NOmega}Omega.txt"

    if os.path.exists(gp_pos_filename) and os.path.exists(gp_neg_filename) and os.path.exists(gp_neutral_filename):
        gp_arrays.gp_pos = np.loadtxt(gp_pos_filename)
        gp_arrays.gp_neg = np.loadtxt(gp_neg_filename)
        gp_arrays.gp_neutral = np.loadtxt(gp_neutral_filename)
        return

    # Generate log-spaced arrays for psi, phi, and Omega
    psi_tab = makelogtab(gp_arrays.psi_min, gp_arrays.psi_max, gp_arrays.Npsi)
    phi_tab = makelogtab(gp_arrays.phi_min, gp_arrays.phi_max, gp_arrays.Nphi)
    Omega_tab = makelogtab(gp_arrays.Omega_min, gp_arrays.Omega_max, gp_arrays.NOmega)

    # Initialize arrays for storing results
    gp_pos = np.zeros((gp_arrays.Npsi, gp_arrays.NOmega))
    gp_neg = np.zeros((gp_arrays.Npsi, gp_arrays.NOmega))
    gp_neutral = np.zeros((gp_arrays.Nphi, gp_arrays.NOmega))

    # --- Compute for charged grains ---
    '''
    for ipsi, psi in enumerate(psi_tab):
        for iOmega, Omega in enumerate(Omega_tab):
            gp_pos[ipsi, iOmega] = little_gp_charged(psi, Omega)
            gp_neg[ipsi, iOmega] = little_gp_charged(-psi, Omega)
    '''
    # Define function for parallel computation
    def compute_pos(ipsi):
        psi = psi_tab[ipsi]
        result = np.zeros(gp_arrays.NOmega)
        for iOmega, Omega in enumerate(Omega_tab):
            result[iOmega] = little_gp_charged(psi, Omega)
        return result
    
    def compute_neg(ipsi):
        psi = psi_tab[ipsi]
        result = np.zeros(gp_arrays.NOmega)
        for iOmega, Omega in enumerate(Omega_tab):
            result[iOmega] = little_gp_charged(-psi, Omega)
        return result
    
    # Parallel computation
    gp_pos = np.array(parallel_map(compute_pos, list(range(gp_arrays.Npsi))))
    gp_neg = np.array(parallel_map(compute_neg, list(range(gp_arrays.Npsi))))
        

    if rank0:
        # Save gp_pos to file
        np.savetxt(gp_pos_filename, gp_pos)

        # Save gp_neg to file    
        np.savetxt(gp_neg_filename, gp_neg)

    # --- Compute for neutral grains ---
    """
    for iphi, phi in enumerate(phi_tab):
        for iOmega, Omega in enumerate(Omega_tab):
            gp_neutral[iphi, iOmega] = little_gp_neutral(phi, Omega)    
    """

    def compute_neutral(iphi):
        phi = phi_tab[iphi]
        result = np.zeros(gp_arrays.NOmega)
        for iOmega, Omega in enumerate(Omega_tab):
            result[iOmega] = little_gp_neutral(phi, Omega)
        return result
    
    gp_neutral = np.array(parallel_map(compute_neutral, list(range(gp_arrays.Nphi))))
    
    if rank0:
        # Save gp_neutral to file
        np.savetxt(gp_neutral_filename, gp_neutral)

    # Store the results
    gp_arrays.gp_pos = gp_pos
    gp_arrays.gp_neg = gp_neg
    gp_arrays.gp_neutral = gp_neutral

    return 

compute_little_gp()
barrier()

Npsi = gp_arrays.Npsi
NOmega = gp_arrays.NOmega
psi_min = gp_arrays.psi_min
psi_max = gp_arrays.psi_max
Omega_min = gp_arrays.Omega_min
Omega_max = gp_arrays.Omega_max

gp_pos = gp_arrays.gp_pos
gp_neg = gp_arrays.gp_neg


def loop_charged(ipsi, psi_arr, NOmega_arr, psi_indices, psi_coeff, Omega_indices, Omega_coeff, gp_pos, gp_neg):
    result = np.zeros(NOmega_arr)
    psi_indices_charged = psi_indices[ipsi]
    psi_coeff_charged = psi_coeff[ipsi]
    Omega_indices_charged = Omega_indices[ipsi, :]
    Omega_coeff_charged = Omega_coeff[ipsi, :]

    if psi_arr[ipsi] > 0:
        result = np.exp(psi_coeff_charged * (Omega_coeff_charged * np.log(gp_pos[psi_indices_charged, Omega_indices_charged]) + 
                        (1 - Omega_coeff_charged) * np.log(gp_pos[psi_indices_charged, Omega_indices_charged + 1])) +
                        (1 - psi_coeff_charged) * (Omega_coeff_charged * np.log(gp_pos[psi_indices_charged + 1, Omega_indices_charged]) + 
                        (1 - Omega_coeff_charged) * np.log(gp_pos[psi_indices_charged + 1, Omega_indices_charged + 1]))
                        )
    elif psi_arr[ipsi] < 0:
        result = np.exp(
            psi_coeff_charged * (Omega_coeff_charged * np.log(gp_neg[psi_indices_charged, Omega_indices_charged]) + 
                             (1 - Omega_coeff_charged) * np.log(gp_neg[psi_indices_charged, Omega_indices_charged + 1])) +
            (1 - psi_coeff_charged) * (Omega_coeff_charged * np.log(gp_neg[psi_indices_charged + 1, Omega_indices_charged]) + 
                               (1 - Omega_coeff_charged) * np.log(gp_neg[psi_indices_charged + 1, Omega_indices_charged + 1]))
            )
    return result


def little_gp_charged_interpol(psi_arr, Omega_arr):
    """
    Parameters:
    - psi_arr: Array of grain potentials
    - Omega_arr: Array of grain rotation frequencies

    Returns:
    - gp_charged: Interpolated values of gp_charged for the input psi_arr and Omega_arr
    should be a 2D array of size (Npsi, NOmega)
    """

    psi_tab = makelogtab(psi_min, psi_max, Npsi)
    Omega_tab = makelogtab(Omega_min, Omega_max, NOmega)


    # Finding the indices and coefficients for psi
    aux = np.floor(Npsi * np.log(np.absolute(psi_arr) / psi_min) / np.log(psi_max / psi_min) - 0.5)
    psi_indices = np.array([np.int64(x) for x in aux])
    psi_indices = np.clip(psi_indices, 0, Npsi - 2)

    psi_coeff = 1 - Npsi * np.log(np.absolute(psi_arr) / psi_tab[psi_indices]) / np.log(psi_max / psi_min)
    psi_coeff = np.clip(psi_coeff, 0, 1)

    # Finding the indices and coefficients for Omega
    aux = np.floor(NOmega * np.log(Omega_arr / Omega_min) / np.log(Omega_max / Omega_min) - 0.5)
    Omega_indices = np.array([np.int64(x) for x in aux])
    Omega_indices = np.clip(Omega_indices, 0, NOmega - 2)

    Omega_coeff = 1 - NOmega * np.log(Omega_arr / Omega_tab[Omega_indices]) / np.log(Omega_max / Omega_min)
    Omega_coeff = np.clip(Omega_coeff, 0, 1)

    # Initializing the result array
    Npsi_arr = np.size(psi_arr)
    NOmega_arr = np.size(Omega_arr)
    gp_charged = np.zeros((Npsi_arr, NOmega_arr))

    # Expand psi_indices and Omega_indices across the appropriate dimensions
    # Multiply psi_indices across the NOmega_arr dimension
    psi_indices = psi_indices.reshape(-1, 1) @ np.ones((1, NOmega_arr), dtype=np.int64)

    # Multiply Omega_indices across the Npsi_arr dimension
    Omega_indices = np.ones((Npsi_arr, 1), dtype=np.int64) @ Omega_indices.reshape(1, -1)

    # Step 4: Expand psi_coeff and Omega_coeff similarly across the dimensions
    # Expand psi_coeff across the NOmega_arr dimension
    psi_coeff = psi_coeff.reshape(-1, 1) @ np.ones((1, NOmega_arr))

    # Expand Omega_coeff across the Npsi_arr dimension
    Omega_coeff = np.ones((Npsi_arr, 1)) @ Omega_coeff.reshape(1, -1)
    
    # gp_charged = np.array(parallel_map(loop_charged, list(range(Npsi_arr))))
    
    # Not to do parallel computation

    for ipsi in range(Npsi_arr):
        gp_charged[ipsi, :] = loop_charged(ipsi, psi_arr, NOmega_arr, psi_indices, psi_coeff, Omega_indices, Omega_coeff, gp_pos, gp_neg)

    return gp_charged # shape: [Npsi_arr, NOmega_arr]


phi_min = gp_arrays.phi_min
phi_max = gp_arrays.phi_max
Nphi = gp_arrays.Nphi
Omega_min = gp_arrays.Omega_min
Omega_max = gp_arrays.Omega_max
NOmega = gp_arrays.NOmega
gp_neutral = gp_arrays.gp_neutral

#@jit
def little_gp_neutral_interpol(phi, Omega_arr):
    """
    Parameters:
    - phi: Grain potential
    - Omega_arr: Array of grain rotation frequencies

    Returns:
    - gp_neu: Interpolated values of gp_neutral for the input phi and Omega_arr
    should be a 1D array of the same size as Omega_arr
    """

    phi_tab = makelogtab(phi_min, phi_max, Nphi)
    Omega_tab = makelogtab(Omega_min, Omega_max, NOmega)

    # Find the index and coefficient for phi
    phi_index = int(np.floor(Nphi * np.log(phi / phi_min) / np.log(phi_max / phi_min) - 0.5))
    #phi_index = np.max((0, np.min((phi_index, Nphi - 2))))  # Ensure phi_index is within bounds
    phi_index = np.maximum(0, np.minimum(phi_index, Nphi - 2))  # Ensure phi_index is within bounds

    # Calculate interpolation coefficient for phi
    phi_coeff = 1 - Nphi * np.log(phi / phi_tab[phi_index]) / np.log(phi_max / phi_min)
    #phi_coeff = np.clip(phi_coeff, 0, 1)  # Clamp phi_coeff between 0 and 1
    phi_coeff = np.maximum(0, phi_coeff)
    phi_coeff = np.minimum(1, phi_coeff)
    
    # Find indices and coefficients for Omega
    # Omega_indices = np.int64(np.floor(NOmega * np.log(Omega_arr / Omega_min) / np.log(Omega_max / Omega_min) - 0.5))
    Omega_indices = np.array([int(x) for x in np.floor(NOmega * np.log(Omega_arr / Omega_min) / np.log(Omega_max / Omega_min) - 0.5)])
    Omega_indices = np.clip(Omega_indices, 0, NOmega - 2)  # Ensure Omega_indices are within bounds
    #Omega_indices = np.maximum(0, np.minimum(Omega_indices, NOmega - 2))  # Ensure Omega_indices are within bounds

    # Calculate interpolation coefficients for Omega
    Omega_coeff = 1 - NOmega * np.log(Omega_arr / Omega_tab[Omega_indices]) / np.log(Omega_max / Omega_min)
    Omega_coeff = np.clip(Omega_coeff, 0, 1)  # Clamp Omega_coeff between 0 and 1
    #Omega_coeff = np.maximum(0, Omega_coeff)
    #Omega_coeff = np.minimum(1, Omega_coeff)

    # Interpolate gp_neutral values for phi
    gp_neu_phi = phi_coeff * gp_neutral[phi_index, :] + (1 - phi_coeff) * gp_neutral[phi_index + 1, :]

    # Return interpolated values for gp_neutral
    return Omega_coeff * gp_neu_phi[Omega_indices] + (1 - Omega_coeff) * gp_neu_phi[Omega_indices + 1]



#@jit
def Gp_per_mu2_averaged(temp, xh, xC, a, beta, fZ, omega_vec):
    """
    Returns G_p^(AHD09)(a, omega)/mu_ip^2 averaged over grain charges.

    Parameters:
    - env: Environment containing T, xh, and xC.
    - a: Grain size.
    - fZ: Grain charge distribution array. 
    - omega: Frequency values.

    Returns:
    - Gp_over_mu2: Averaged G_p divided by mu_ip^2, as an array of the same dimension as omega. shape: [Nomega]
    """
   
    acx_val = acx(a, beta)

    Nomega = np.size(omega_vec)

    Zg_arr = fZ[0, :]
    fZ_arr = fZ[1, :] 

    little_gp_H = np.zeros(Nomega)
    little_gp_C = np.zeros(Nomega)

    Omega_H = np.sqrt(mp / (2 * k * temp)) * acx_val * omega_vec
    Omega_C = np.sqrt(12) * Omega_H

    # Neutral grain contribution
    phi = np.sqrt(2 / (acx_val * k * temp)) * q
    little_gp_H = fZ[1, 0] * little_gp_neutral_interpol(phi, Omega_H) # shape: [Nomega]
    little_gp_C = fZ[1, 0] * little_gp_neutral_interpol(phi, Omega_C)

    # Charged grain contribution
    ind_charged = np.where(Zg_arr != 0)[0]
    if len(ind_charged) > 0:
        psi_arr = Zg_arr[ind_charged] * q**2 / (acx_val * k * temp)
        #little_gp_H += np.matmul(little_gp_charged_interpol(psi_arr, Omega_H).T, fZ_arr[ind_charged])
        little_gp_H += little_gp_charged_interpol(psi_arr, Omega_H).T @ fZ_arr[ind_charged]
        #little_gp_C += np.matmul(little_gp_charged_interpol(psi_arr, Omega_C).T, fZ_arr[ind_charged])
        little_gp_C += little_gp_charged_interpol(psi_arr, Omega_C).T @ fZ_arr[ind_charged]

    Gp_over_mu2 = (q / (acx_val**2 * k * temp))**2 * (xh * little_gp_H + xC * np.sqrt(12) * little_gp_C)

    return Gp_over_mu2


def FGp_averaged(env, a, beta, fZ, omega_vec, mu_ip, mu_op, tumbling=True):
    """
    Returns a structure {Fp, Gp} with each element as an array of dimensions [Nomega, Nmu].
    (Averaged over grain charges.)
    If tumbling is True, disklike tumbling grain is used.

    Parameters:
    - env: Environment containing T, xh, and xC.
    - a: Grain size.
    - fZ: Grain charge distribution array.
    - omega_vec: Frequency values.
    - mu_ip, mu_op: Arrays of internal and external dipole moments.
    - tumbling: Boolean, whether the grain is tumbling.

    Returns:
    - A dictionary with Fp and Gp arrays. Shape: [Nomega, Nmu]
    """
    mu_ip_2 = np.array(mu_ip)**2
    mu_op_2 = np.array(mu_op)**2
    temp, xh, xC = env['T'], env['xh'], env['xC']
    Gp_op = 2/3 * Gp_per_mu2_averaged(temp, xh, xC, a, beta, fZ, 2 * omega_vec) 
    Nomegas = np.size(omega_vec)

    if beta == 0: # Spherical grain
        tumbling = False

    if tumbling:
        # Disklike tumbling grain 
        ## Corr: omega could be a vector
        def Gp_Fp_vals(omega_ind):
            omega = omega_vec[omega_ind]
            cos_theta_list = maketab(-1, 1, 21)
            Omega_arr = np.absolute(omega  * (1 - beta * cos_theta_list) / (1+beta) )
            weight_Gp = (1 - cos_theta_list)**2 / 4
            weight_Fp = - weight_Gp * cos_theta_list
            aux_result = Gp_per_mu2_averaged(temp, xh, xC, a, beta, fZ, Omega_arr)
            result_Gp = 2*np.average(aux_result*weight_Gp) *  mu_ip_2 + Gp_op[omega_ind] * mu_op_2 
            result_Fp = result_Gp/(1+beta) + (beta/(1+beta)) * 2*np.average(aux_result*weight_Fp) *  mu_ip_2 
            return result_Gp, result_Fp
        
        Gp, Fp = [], []

        for ind in np.arange(Nomegas):
            Gp_vals, Fp_vals = Gp_Fp_vals(ind)
            Gp.append(Gp_vals)
            Fp.append(Fp_vals)

        Gp = np.array(Gp).reshape(Nomegas, -1)
        Fp = np.array(Fp).reshape(Nomegas, -1)
       
        return {'Fp': Fp, 'Gp': Gp}

    # Standard spherical grain with K = J
    Gp = Gp_per_mu2_averaged(temp, xh, xC, a, beta, fZ, omega_vec)
    Gp = np.matmul(Gp.reshape(-1,1), mu_ip_2.reshape(1,-1))
    Fp = Gp

    return {'Fp': Fp, 'Gp': Gp}





from .Grain import acx, asurf
from .charge_dist import refr_indices, hnu_pdt, hnu_pet, Y, Qabs, sigma_pdt, nu_uisrf, E_min, JPEisrf
from .util import DX_over_X, makelogtab, cgsconst
import numpy as np


eV = cgsconst.eV
mp = cgsconst.mp
me = cgsconst.me
q = cgsconst.q
k = cgsconst.k
c = cgsconst.c
hbar = cgsconst.h / (2 * np.pi)

hnu_tab = refr_indices.hnu_tab
la_tab = refr_indices.la_tab   

# Define the GH2 function
def GH2(env, a, beta):
    """
    Excitation rate due to random H2 formation.
    
    Parameters:
    - env: an object or dictionary containing environment properties T, nh, y, gamma
    - a, beta: parameter for which `acx(a, beta)` will be calculated

    Returns:
    - Excitation rate (GH2)
    """
    
    # Extract environment variables
    T = env['T']
    y = env['y']
    gamma = env['gamma']
    
    # Calculate ax from acx(a)
    ax = acx(a, beta)

    # Define Ef (formation energy) and JJplus1 (quantum number term)
    Ef = 0.2 * eV
    JJplus1 = 1e2  # Equivalent to 1d2 in IDL

    # Calculate GH2 according to the formula
    GH2_value = (gamma / 4) * (1 - y) * Ef / (k * T) * (1 + JJplus1 * hbar**2 / (2 * mp * Ef * ax**2))
    
    return GH2_value

# Function FGpeZ
def FGpeZ(env, a, beta, Z):
    """
    Calculate Fpe and Gpe for a given environment, grain size 'a', and charge 'Z'.
    
    Parameters:
    - env: Dictionary containing environment parameters like T, nh, Chi.
    - a: Grain size.
    - Z: Charge state.

    Returns:
    - A dictionary with Fpe and Gpe values.
    """

    # Environment variables
    T = env['T']
    nh = env['nh']
    Chi = env['Chi']
    asurf_val = asurf(a, beta)  

    # Get Jpeisrf 
    aux1, aux2 = JPEisrf(a)
    Jpepos = Chi * aux1
    Jpeneg = Chi * aux2

    # --- Gpe ---
    hnu_pet_val = max(1e-5, hnu_pet(Z, a)) 
    hnu_pdt_val = max(1e-5, hnu_pdt(Z, a)) 
    Nnu = 500
    hnu_pet_tab = makelogtab(hnu_pet_val, 13.6, Nnu) 
    Dnu_over_nu_pet = DX_over_X(hnu_pet_val, 13.6, Nnu)  
    hnu_pdt_tab = makelogtab(hnu_pdt_val, 13.6, Nnu)
    Dnu_over_nu_pdt = DX_over_X(hnu_pdt_val, 13.6, Nnu)

    Ytab = np.interp(np.log(hnu_pet_tab), np.log(hnu_tab), Y(Z,a))
    Qtab = Qabs(a, Z, hnu_pet_tab)  

    if Z >= 0:
        E_low = -(Z + 1) * q**2 / a
        E_high = (hnu_pet_tab - hnu_pet_val) * eV
        Epe = 0.5 * E_high * (E_high - 2 * E_low) / (E_high - 3 * E_low)  # In erg
        second_term = 0
        third_term = Jpepos[int(Z)] * (Z + 1) * q**2 / asurf_val
    else:
        E_min_val = E_min(Z, a) 
        E_low = E_min_val * eV
        E_high = (E_min_val + hnu_pet_tab - hnu_pet_val) * eV
        Epe = 0.5 * (E_high + E_low)  # In erg
        second_term = Dnu_over_nu_pdt * np.sum(sigma_pdt(hnu_pdt_tab, Z, a) * nu_uisrf(hnu_pdt_tab) / eV / hnu_pdt_tab 
                                               * (hnu_pdt_tab - hnu_pdt_val + E_min_val))
        third_term = Jpeneg[int(-Z)] * (Z + 1) * q**2 / asurf_val

    first_term = c * np.pi * a**2 * Dnu_over_nu_pet * np.sum(Ytab * Qtab * nu_uisrf(hnu_pet_tab) / eV / hnu_pet_tab * Epe)

    Gpe = me / (4 * nh * np.sqrt(8 * np.pi * mp * k * T) * asurf_val**2 * k * T) * (first_term + second_term + third_term)

    # --- Fpe ---
    if Z >= 0:
        Jpe = Jpepos[int(Z)]
    else:
        Jpe = Jpeneg[int(-Z)]

    Fpe = me / mp * Jpe / (2 * np.pi * asurf_val**2 * nh * np.sqrt(2 * k * T / (np.pi * mp)))

    return {'Fpe': Fpe, 'Gpe': Gpe}


def FGpe_averaged(env, a, beta, fZ):
    """
    Computes the averaged values of Fpe and Gpe over grain charges.

    Parameters:
    - env: Environment parameters.
    - a: Grain size.
    - fZ: Grain charge distribution, a 2D array where fZ[0, :] contains grain charges 
          and fZ[1, :] contains their corresponding probabilities.

    Returns:
    - A dictionary with averaged Fpe and Gpe values.
    """
    NZ = fZ.shape[1]  # Number of grain charges
    Fpe = 0.0
    Gpe = 0.0

    # Loop over all grain charge states
    for i in range(NZ):
        FGpe = FGpeZ(env, a, beta, fZ[0, i])  # Get the Fpe and Gpe for charge state Zg
        Fpe += fZ[1, i] * FGpe['Fpe']   # Weighted Fpe
        Gpe += fZ[1, i] * FGpe['Gpe']   # Weighted Gpe

    return {'Fpe': Fpe, 'Gpe': Gpe}



from ..util import cgsconst, makelogtab
from .grain_properties import N_C, N_H
from .emissivity import emissivity, free_free
import numpy as np

pi = np.pi
c = cgsconst.c
q = cgsconst.q
k = cgsconst.k
mp = cgsconst.mp
debye = cgsconst.debye
eV = cgsconst.eV


def SPDUST(environment, tumbling=True, output_file=None, min_freq=None, max_freq=None, n_freq=None, Ndipole=None, freefree=False):
    
    # Check the environment structure for required parameters
    if 'dipole' not in environment and 'beta' not in environment:
        print("Please specify the dipole moment or beta.")
        return
    
    # Determine the dipole moment
    if 'beta' in environment:
        beta0 = environment['beta'] * debye
        mu_1d_7 = np.sqrt(N_C(1e-7) + N_H(1e-7)) * environment['beta']
    else:
        mu_1d_7 = environment['dipole']
        beta0 = mu_1d_7 / np.sqrt(N_C(1e-7) + N_H(1e-7)) * debye

    # Check for grain size distribution parameters
    if 'line' not in environment:
        print("Please specify the grain size distribution parameters (Weingartner & Draine, 2001a).")
        return

    # Number of dipole moments
    Ndip = 20
    if Ndipole is not None:
        Ndip = Ndipole

    # Set in-plane moment ratio
    ip = 2 / 3
    if 'inplane' in environment:
        print(f"Assuming that <mu_ip^2>/<mu^2> = {environment['inplane']} for disklike grains")
        ip = environment['inplane']

    # Frequency settings
    GHz = 1e9
    numin = 0.5 * GHz
    numax = 500 * GHz
    Nnu = 200
    if min_freq is not None:
        numin = min_freq * GHz
    if max_freq is not None:
        numax = max_freq * GHz
    if n_freq is not None:
        Nnu = n_freq

    nu_tab = makelogtab(numin, numax, Nnu)

    # Debugging
    print(f"mu(1E-7 cm) = {mu_1d_7} debye (beta = {beta0 / debye} debye)")

    # Calculate emissivity
    if tumbling:
        jnu_per_H = emissivity(environment, beta0, ip, Ndip, nu_tab, tumbling=True)
    else:
        print("Assuming that disklike grains spin around their axis of greatest inertia")
        jnu_per_H = emissivity(environment, beta0, ip, Ndip, nu_tab, tumbling=False)
        

    # Handle free-free emission if requested
    Jy = 1e-23
    result = np.zeros((3 if freefree else 2, Nnu))
    if freefree:
        result[2, :] = free_free(environment, nu_tab) / Jy
    result[0, :] = nu_tab / GHz
    result[1, :] = jnu_per_H / Jy

    # Write output to file
    if output_file is not None:        
        with open(output_file, 'w') as f:
            f.write('#=========================== SPDUST.2 ===============================\n')
            f.write('#    Rotational emission from a population of spinning dust grains,\n')
            f.write('#    as described by Ali-Haimoud, Hirata & Dickinson, 2009, MNRAS, 395, 1055\n')
            f.write('#    and in Silsbee, Ali-Haimoud & Hirata, 2010\n')
            f.write(f'#    nH = {environment["nh"]} cm^-3\n')
            f.write(f'#    T = {environment["T"]} K\n')
            f.write(f'#    Chi = {environment["Chi"]}\n')
            f.write(f'#    xh = {environment["xh"]}\n')
            f.write(f'#    xC = {environment["xC"]}\n')
            f.write(f'#    mu(1E-7 cm) = {mu_1d_7} debye (beta = {beta0 / debye} debye)\n')
            if tumbling:
                f.write('#    Disklike grains are randomly oriented with respect to angular momentum.\n')
            else:
                f.write('#    Disklike grains spin around their axis of greatest inertia\n')
            f.write('#=====================================================================\n')
            if freefree:
                f.write('#nu(GHz)       j_nu/nH(Jy sr-1 cm2/H)     j_nu/nH (free-free) (Jy sr-1 cm2/H)\n')
            else:
                f.write('#nu(GHz)       j_nu/nH(Jy sr-1 cm2/H)\n')
            np.savetxt(f, result.T, fmt='%12.6e') # Columns are nu, jnu_per_H, jnu_per_H_freefree
    
    return result # shape (2, Nnu) or (3, Nnu) if freefree is True; rows are nu, jnu_per_H, jnu_per_H_freefree





import numpy as np
from scipy.special import erf
from .. import SpDust_data_dir
from ..util import cgsconst
import os


mp = cgsconst.mp
q = cgsconst.q
pi = np.pi

class grainparams:
    # Stores some grain geometrical properties
    a2 = 6e-8          # radius under which grains are disklike
    d = 3.35e-8        # disk thickness (graphite interlayer separation)
    rho = 2.24         # carbon density in g/cm^3
    epsilon = 0.01     # charge centroid displacement

rho = grainparams.rho
d = grainparams.d
a2 = grainparams.a2
epsilon = grainparams.epsilon


# Number of carbon atoms in a grain
#@njit
def N_C(a):
    return int(np.floor(4 * pi / 3 * a**3 * rho / (12 * mp)) + 1)

# Number of hydrogen atoms in a grain
#@njit
def N_H(a):
    Nc = N_C(a)
    if Nc < 25:
        result = np.floor(0.5 * Nc + 0.5)
    elif Nc < 100:
        result = np.floor(2.5 * np.sqrt(Nc) + 0.5)
    else:
        result = np.floor(0.25 * Nc + 0.5)
    return int(result)

# Largest moment of inertia
#@njit
def Inertia(a):
    Mass = (12 * N_C(a) + N_H(a)) * mp
    Isphere = 0.4 * Mass * a**2

    if a <= a2:
        return 5 / 3 * a / d * Isphere   # DL98b (A8)
    else:
        return Isphere

# Surface-equivalent radius
#@njit
def asurf(a):
    if a <= a2:
        b = np.sqrt(4 / 3 * a**3 / d)
        return np.sqrt(b**2 / 2 + b * d / 2)
    else:
        return a

# Cylindrical excitation equivalent radius
#@njit
def acx(a):
    if a <= a2:
        R = np.sqrt(4 / 3 * a**3 / d)
        return (3 / 8)**0.25 * R
    else:
        return a

# Root mean square dipole moment
#@njit
def rms_dipole(a, Z2, beta):
    muZ = epsilon * np.sqrt(Z2) * q * acx(a)
    N_at = N_C(a) + N_H(a)
    return np.sqrt(N_at * beta**2 + muZ**2)




# Define the path to the size distribution file
size_dist_file = os.path.join(SpDust_data_dir, 'sizedists_table1.out')

class size_dist_arrays:
    bc1e5_tab, alpha_tab, beta_tab, at_tab, ac_tab, C_tab = \
    np.loadtxt(size_dist_file, usecols=(1, 3, 4, 5, 6, 7), unpack=True, comments=';')

class size_params():
    bc, alpha_g, beta_g, at_g, ac_g, C_g = None, None, None, None, None, None

    def __call__(self, line):
        self.bc = size_dist_arrays.bc1e5_tab[line] * 1e-5
        self.alpha_g = size_dist_arrays.alpha_tab[line]
        self.beta_g = size_dist_arrays.beta_tab[line]
        self.at_g = size_dist_arrays.at_tab[line] * 1e-4
        self.ac_g = size_dist_arrays.ac_tab[line] * 1e-4
        self.C_g = size_dist_arrays.C_tab[line]
        pass

# Grain size distribution
def size_dist(a, line):
    '''
    Grain size distribution, using Weingartner & Draine, 2001a prescription. 
    The line of their table 1 is given by the user in param_file.
    '''
    size_dist_params = size_params()
    size_dist_params(line)

    # Unpack size distribution parameters
    bc = size_dist_params.bc
    alpha_g = size_dist_params.alpha_g
    beta_g = size_dist_params.beta_g
    at_g = size_dist_params.at_g
    ac_g = size_dist_params.ac_g
    C_g = size_dist_params.C_g

    rho = grainparams.rho
    # Lognormal populations parameters
    mc = 12 * mp
    bci = np.array([0.75, 0.25]) * bc
    a0i = np.array([3.5, 30.]) * 1e-8
    sigma = 0.4
    amin = 3.5e-8

    Bi = 3 / (2 * np.pi)**1.5 * np.exp(-4.5 * sigma**2) / (
        rho * a0i**3 * sigma) * bci * mc / (1 + erf(3 * sigma / np.sqrt(2) + np.log(a0i / amin) / (sigma * np.sqrt(2))))
    D_a = np.sum(Bi / a * np.exp(-0.5 * (np.log(a / a0i) / sigma)**2))

    if beta_g >= 0:
        F_a = 1 + beta_g * a / at_g
    else:
        F_a = 1 / (1 - beta_g * a / at_g)

    cutoff = 1
    if a > at_g:
        cutoff = np.exp(-((a - at_g) / ac_g)**3)

    return D_a + C_g / a * (a / at_g)**alpha_g * F_a * cutoff



from .. import SpDust_data_dir
from ..util import cgsconst, DX_over_X, maketab, makelogtab 
from .grain_properties import acx, Inertia, grainparams, rms_dipole, size_dist
from .charge_dist import charge_dist
from .infrared import FGIR_averaged
from .collisions import Tev_effective, FGn_averaged, FGi_averaged
from .plasmadrag import FGp_averaged
from .H2_photoemission import GH2, FGpe_averaged

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d



pi = np.pi
c = cgsconst.c
h = cgsconst.h
k = cgsconst.k
eV = cgsconst.eV
me = cgsconst.me
q = cgsconst.q
mp = cgsconst.mp

# Function to calculate the characteristic damping time through collisions with neutral H atoms
def tau_H(env, a):
    """
    Returns the characteristic damping time through collisions with neutral H atoms.
    This corresponds to Eq. (24) in AHD09.

    Parameters:
    - env: Environment parameters (temperature T, hydrogen density nh).
    - a: Grain size.

    Returns:
    - Damping time tau_H (in seconds).
    """
    T = env['T']  # temperature in K
    nh = env['nh']  # hydrogen density in cm^-3
    acx_val = acx(a)  # characteristic length scale
    Inertia_val = Inertia(a)  # grain's moment of inertia

    # Calculate tau_H using the given equation
    return 1.0 / (nh * mp * np.sqrt(2 * k * T / (pi * mp)) * 4 * pi * acx_val**4 / (3 * Inertia_val))

# Function to calculate the inverse of the characteristic damping time through electric dipole radiation
def tau_ed_inv(env, a, mu_ip, mu_op, tumbling=True, corrected=False):
    """
    Returns the inverse of the characteristic damping time through electric dipole radiation.
    This corresponds to Eq. (29) in AHD09 and Eq. (47) in SAH10 for tumbling disklike grains.

    Parameters:
    - env: Environment parameters (temperature T).
    - a: Grain size.
    - mu_ip: Electric dipole moment in-plane.
    - mu_op: Electric dipole moment out-of-plane.
    - tumbling: Boolean, if True, use the tumbling disk model.

    Returns:
    - Inverse of the damping time (in seconds^-1).
    """
    T = env['T']  # temperature in K
    Inertia_val = Inertia(a)  # grain's moment of inertia

    if corrected:
        if tumbling and a < grainparams.a2:
            # Tumbling disklike grains, SAH10 Eq. (47)
            return (3 * k * T * (82 / 45 * mu_ip**2 + 32 / 9 * mu_op**2)) / (Inertia_val**2 * c**3)
        else:
            # Spherical grains or disklike grains with K = J, AHD09 Eq. (29)
            return (2 * k * T * mu_ip**2) / (Inertia_val**2 * c**3)
    else: # Use the old  IDL version - without the correction
        if tumbling:
            # Tumbling disklike grains, SAH10 Eq. (47)
            return (3 * k * T * (82 / 45 * mu_ip**2 + 32 / 9 * mu_op**2)) / (Inertia_val**2 * c**3)
        else:
            # Spherical grains or disklike grains with K = J, AHD09 Eq. (29)
            return (2 * k * T * mu_ip**2) / (Inertia_val**2 * c**3)

    
def f_rot(env, a, fZ, mu_ip, mu_op, tumbling=True):
    """
    Returns the rotational distribution function f_a normalized such that:
    ∫ f_a(omega) * 4 * pi * omega^2 d(omega) = 1.
    The function is calculated at frequencies centered on the approximate peak frequency
    of the emitted power. The output is a dictionary {omega, f_a} containing:
    - omega: array of frequencies.
    - f_a: rotational distribution function (2D array with dimensions [Nmu, Nomega]).

    Parameters:
    - env: Environmental parameters (T, nh).
    - a: Grain size.
    - ia: Index or identifier for intermediate value saving (optional).
    - fZ: Distribution of grain charges.
    - mu_ip: Electric dipole moment in-plane.
    - mu_op: Electric dipole moment out-of-plane.
    - tumbling: Boolean, whether tumbling disklike grains are considered.

    Returns:
    - Dictionary containing omega and f_a arrays.
    """
    Nomega = 1000  # Number of omega values for which the function is calculated
    Nmu = np.size(mu_ip)  # Number of dipole moments

    Tval = env['T']
    Inertia_val = Inertia(a)  # Grain's moment of inertia

    # Characteristic timescales
    tau_H_val = tau_H(env, a)
    tau_ed_inv_val = tau_ed_inv(env, a, mu_ip, mu_op, tumbling=tumbling)
    # Evaporation temperature
    Tev_val = Tev_effective(env, a)

    # F's and G's (except for plasma drag)
    FGn = FGn_averaged(env, a, Tev_val, fZ, tumbling=tumbling)
    Fn = FGn['Fn']
    Gn = FGn['Gn']

    mu_tot = np.sqrt(mu_ip**2 + mu_op**2)
    FGi = FGi_averaged(env, a, Tev_val, mu_tot, fZ)
    Fi = FGi['Fi']
    Gi = FGi['Gi']

    FGIR = FGIR_averaged(env, a, fZ)
    FIR = FGIR['FIR']
    GIR = FGIR['GIR']

    FGpe = FGpe_averaged(env, a, fZ)
    Fpe = FGpe['Fpe']
    Gpe = FGpe['Gpe']

    GH2_val = GH2(env, a)

    # Array of omegas around the approximate peak
    omega_peak_th = np.sqrt(6 * k * Tval / Inertia_val)  # Peak of spectrum if thermal rotation

    # Peak frequency for the lowest and highest values of mu_ip, mu_op
    FGp = FGp_averaged(env, a, fZ, omega_peak_th, [np.min(mu_ip), np.max(mu_ip)], [np.min(mu_op), np.max(mu_op)], tumbling=tumbling)
    Fp_th = FGp['Fp']
    Gp_th = FGp['Gp']
    

    F_low = Fn + np.min(Fi) + FIR + Fpe + np.min(Fp_th)
    G_low = Gn + np.min(Gi) + GIR + Gpe + GH2_val + np.min(Gp_th)
    xi_low = 8 * G_low / F_low**2 * tau_H_val * np.min(tau_ed_inv_val)

    F_high = Fn + np.max(Fi) + FIR + Fpe + np.max(Fp_th)
    G_high = Gn + np.max(Gi) + GIR + Gpe + GH2_val + np.max(Gp_th)
    xi_high = 8 * G_high / F_high**2 * tau_H_val * np.max(tau_ed_inv_val)

    omega_peak_low = omega_peak_th * np.sqrt(2 * G_low / F_low / (1 + np.sqrt(1 + xi_low)))
    omega_peak_high = omega_peak_th * np.sqrt(2 * G_high / F_high / (1 + np.sqrt(1 + xi_high)))


    # Array omega
    omega_min = 5e-3 * np.min((omega_peak_low, omega_peak_high))
    omega_max = 6 * np.max((omega_peak_low, omega_peak_high))
    omega = makelogtab(omega_min, omega_max, Nomega) 
    Dln_omega = DX_over_X(omega_min, omega_max, Nomega) 

    # Fp(omega), Gp(omega)
    FGp = FGp_averaged(env, a, fZ, omega, mu_ip, mu_op, tumbling=tumbling)
    Fp = FGp['Fp'] # shape (Nomega, Nmu)
    Gp = FGp['Gp']

    # Rotational distribution function, AHD09 Eq.(33)
    f_a = np.zeros((Nomega, Nmu))

    F = Fn + FIR + Fpe + np.matmul(np.ones((Nomega, 1)), Fi.reshape(1, -1)) + Fp
    G = Gn + GIR + Gpe + GH2_val + np.matmul(np.ones((Nomega, 1)), Gi.reshape(1, -1)) + Gp
    tau_ed_inv_matrix = np.matmul(np.ones((Nomega, 1)), tau_ed_inv_val.reshape(1, -1))
    omega_tab = np.matmul(omega.reshape(-1, 1), np.ones((1, Nmu)))

    X = Inertia_val * omega_tab**2 / (k * Tval)
    integrand = F / G * X + tau_H_val / (3 * G) * tau_ed_inv_matrix * X**2
    exponent = np.cumsum(integrand, axis=0) * Dln_omega
    norm = 4 * pi * np.sum(omega_tab**3 * np.exp(-exponent), axis=0) * Dln_omega
    norm = np.matmul(np.ones((Nomega, 1)), norm.reshape(1, -1))

    f_a = 1 / norm * np.exp(-exponent)

    return {'omega': omega, 'f_a': f_a.T}  # Transpose for consistency with dimensions [Nmu, Nomega]


def mu2_fa(env, a, fZ, mu_rms, ip, Ndipole, tumbling=True):
    """
    Returns a [3, Nomega] array:
    [omega, <mu_ip^2 fa(omega)>/<mu_ip^2>, <mu_op^2 fa(omega)>/<mu_op^2>].
    
    Parameters:
    - env: Environment parameters (such as temperature).
    - a: Grain size.
    - ia: Index for saving intermediate results.
    - fZ: Distribution of grain charges.
    - mu_rms: Root mean square dipole moment.
    - ip: Fraction of in-plane dipole moment.
    - Ndipole: Number of dipole values used in averaging.
    - tumbling: Boolean to indicate whether the grains are tumbling disklike grains.

    Returns:
    - A numpy array [3, Nomega] containing omega, <mu_ip^2 fa(omega)>/<mu_ip^2>, and <mu_op^2 fa(omega)>/<mu_op^2>.
    """

    op = 1.0 - ip  # Out-of-plane dipole moment fraction

    if Ndipole == 1:  # If no averaging over dipoles
        f_rot_data = f_rot(env, a, fZ, mu_rms * np.sqrt(ip), mu_rms * np.sqrt(op), tumbling=tumbling)
        omega = f_rot_data['omega']
        f_a = f_rot_data['f_a']
        Nomega = len(omega)
        mu_ip2_fa = f_a
        mu_op2_fa = f_a

    else:
        # Set up for averaging
        xmin = 5e-3
        xmed = 0.5
        xmax = 5.0
        aux_Nd = int(Ndipole/2)
        x_tab = np.concatenate((makelogtab(xmin, xmed, aux_Nd), 
                                maketab(xmed, xmax, aux_Nd)))
        Dx_tab = np.concatenate((DX_over_X(xmin, xmed, aux_Nd) * makelogtab(xmin, xmed, aux_Nd), 
                                 (xmax - xmed)/aux_Nd + np.zeros(aux_Nd)))

        if a < grainparams.a2:  # Disk-like grains, need 2D Gaussian
            mu_ip = np.sqrt(ip) * mu_rms * np.outer(x_tab, np.ones(Ndipole))
            mu_op = np.sqrt(op) * mu_rms * np.outer(np.ones(Ndipole), x_tab)
            Dmu_ip = np.outer(Dx_tab, np.ones(Ndipole))
            Dmu_op = Dmu_ip.T

            # Probability calculation for dipoles
            if ip == 0.0:
                Proba = np.exp(-0.5 * mu_op**2 / mu_rms**2) * Dmu_op
            elif op == 0.0:
                Proba = mu_ip / mu_rms * np.exp(-mu_ip**2 / mu_rms**2) * Dmu_ip
            else:
                Proba = (mu_ip / mu_rms * np.exp(-mu_ip**2 / (ip * mu_rms**2)) * 
                         np.exp(-0.5 * mu_op**2 / (op * mu_rms**2)) * Dmu_ip * Dmu_op)

            Proba = Proba / np.sum(Proba)
            # Flatten 2D arrays to 1D
            mu_ip = mu_ip.flatten()
            mu_op = mu_op.flatten()
            Proba = Proba.flatten()

            f_rot_data = f_rot(env, a, fZ, mu_ip, mu_op, tumbling=tumbling)
            omega = f_rot_data['omega']
            f_a = f_rot_data['f_a']
            Nomega = np.size(omega)

            Proba = np.outer(Proba, np.ones(Nomega))
            mu_ip = np.outer(mu_ip, np.ones(Nomega))
            mu_op = np.outer(mu_op, np.ones(Nomega))

            # Calculate <mu_ip^2 fa(omega)> and <mu_op^2 fa(omega)>
            mu_ip2_fa = np.sum((mu_ip**2 / mu_rms**2) * Proba * f_a, axis=0)
            mu_op2_fa = np.sum((mu_op**2 / mu_rms**2) * Proba * f_a, axis=0)

            if ip != 0.0 and op != 0.0:
                mu_ip2_fa /= ip
                mu_op2_fa /= op

        else:  # Spherical grains, average over grain orientation first
            Proba = x_tab**2 * np.exp(-1.5 * x_tab**2) * Dx_tab
            Proba = Proba / np.sum(Proba)

            f_rot_data = f_rot(env, a, fZ, np.sqrt(2/3) * mu_rms * x_tab, mu_rms / np.sqrt(3) * x_tab, tumbling=tumbling)
            omega = f_rot_data['omega']
            f_a = f_rot_data['f_a']
            Nomega = np.size(omega)
            Proba = np.outer(Proba, np.ones(Nomega))
            x_tab = np.outer(x_tab, np.ones(Nomega))

            mu2_fa_aux = np.sum(x_tab**2 * Proba * f_a, axis=0)
            mu_ip2_fa = mu2_fa_aux 
            mu_op2_fa = mu2_fa_aux 

    # Prepare the result as a [3, Nomega] array
    result = np.zeros((3, Nomega))
    result[0, :] = omega
    result[1, :] = mu_ip2_fa
    result[2, :] = mu_op2_fa

    # Save the intermediate results (optional)
    # print('Saving results...')
    # np.savez(f'{ia}_mu_fa.npz', omega=omega, mu_ip2_fa=mu_ip2_fa, mu_op2_fa=mu_op2_fa, mu_rms=mu_rms, ip=ip)

    return result


def fa_ip_fa_op(omega, mu_ip2_fa, mu_op2_fa):
    """
    Calculates fa_ip and fa_op functions for tumbling grains.

    Parameters:
    - omega: Array of angular frequencies (logarithmically spaced).
    - mu_ip2_fa: Array <mu_ip^2 fa>/<mu_ip^2>.
    - mu_op2_fa: Array <mu_op^2 fa>/<mu_op^2>.

    Returns:
    - A 2D array of shape [3, count] containing:
      - omega values where omega/3 is possible,
      - fa_ip values,
      - fa_op values.
    """

    Dln_omega = np.sqrt(omega[1] / omega[0]) - np.sqrt(omega[0] / omega[1])

    # Find indices where omega/3 is possible
    ind = np.where(omega / 3 > np.min(omega))[0]

    # Interpolate fa_op at omega/2 using cubic spline interpolation
    interp_op = interp1d(np.log(omega), mu_op2_fa, kind='cubic', fill_value="extrapolate")
    fa_op = interp_op(np.log(omega[ind] / 2)) / 8.0

    # Cumulative integrals for mu_ip2_fa
    int0 = np.cumsum(mu_ip2_fa[::-1])[::-1] * Dln_omega
    int1 = np.cumsum((omega * mu_ip2_fa)[::-1])[::-1] / omega * Dln_omega
    int2 = np.cumsum((omega**2 * mu_ip2_fa)[::-1])[::-1] / omega**2 * Dln_omega

    # Interpolate these integrals at omega/3
    interp_int0 = interp1d(np.log(omega), int0, kind='cubic', fill_value="extrapolate")
    interp_int1 = interp1d(np.log(omega), int1, kind='cubic', fill_value="extrapolate")
    interp_int2 = interp1d(np.log(omega), int2, kind='cubic', fill_value="extrapolate")


    int0_low = interp_int0(np.log(omega[ind] / 3))
    int1_low = interp_int1(np.log(omega[ind] / 3))
    int2_low = interp_int2(np.log(omega[ind] / 3))

    # Calculate fa_ip
    fa_ip = (int0_low - 2 * int1_low + int2_low 
             - 3 * int0[ind] + 6 * int1[ind] - 7 * int2[ind]) / 4.0

    # Prepare the result as a [3, count] array
    result = np.zeros((3, len(ind)))
    result[0, :] = omega[ind]
    result[1, :] = fa_ip
    result[2, :] = fa_op

    return result


def dP_dnu_dOmega(env, a, beta, ip, Ndipole, tumbling=True):
    """
    Returns the power radiated by a grain of radius `a`, per frequency interval, per steradian.
    The result is an array Power_per_grain[2, Nnu] such that:
        Power_per_grain[0,*] = nu
        Power_per_grain[1,*] = dP/dnu/dOmega
    """
    
    op = 1.0 - ip
    
    # Charge distribution and rms dipole moment
    fZ = charge_dist(env, a)
    Z2 = np.sum(fZ[0, :]**2 * fZ[1, :])
    mu_rms = rms_dipole(a, Z2, beta)
    
    # Calculate rotational distribution for in-plane and out-of-plane dipole moments
    mu2_fa_aux  = mu2_fa(env, a, fZ, mu_rms, ip, Ndipole, tumbling=tumbling)
    omega = mu2_fa_aux[0, :]
    mu_ip2_fa = mu2_fa_aux[1, :]
    mu_op2_fa = mu2_fa_aux[2, :]
    
    # Tumbling case for disklike grains
    if tumbling:
        fa_ip_fa_op_aux = fa_ip_fa_op(omega, mu_ip2_fa, mu_op2_fa)
        nu = fa_ip_fa_op_aux[0, :] / (2.0 * pi)
        fa_ip = fa_ip_fa_op_aux[1, :]
        fa_op = fa_ip_fa_op_aux[2, :]
        dPdnudOmega = 2.0 / (3.0 * c**3) * (2.0 * np.pi * nu)**6 * 2.0 * np.pi * mu_rms**2 * (ip * fa_ip + 2.0/3.0 * op * fa_op)
    else:
        # Spherical / disklike with K = J case
        nu = omega / (2.0 * pi)
        dPdnudOmega = 2.0 / (3.0 * c**3) * omega**6 * 2.0 * np.pi * mu_rms**2 * ip * mu_ip2_fa
    
    # Keep only non-zero values
    non_zero_indices = np.where(dPdnudOmega > 0)[0]
    if len(non_zero_indices) > 0:
        non_zero_power = np.zeros((2, len(non_zero_indices)))
        non_zero_power[0, :] = nu[non_zero_indices]
        non_zero_power[1, :] = dPdnudOmega[non_zero_indices]
        return non_zero_power
    else:
        print(f"Power vanishes for a = {a}")
        return np.array([0.0, 0.0])

def emissivity(env, beta, ip, Ndipole, nu_tab, tumbling=True):
    """
    Parameters:
    - env: Environment parameters.


    Returns j_nu/nH in cgs units (ergs/s/sr/H atom) for the given environment.
    """
    
    # Grain parameters
    a_min = 3.5e-8
    a_max = 3.5e-7  # Large grains do not contribute near the peak
    Na = 30
    a_tab = makelogtab(a_min, a_max, Na)
    Da_over_a = DX_over_X(a_min, a_max, Na)
    
    # Initialize emissivity array
    emiss = np.zeros(len(nu_tab))
    line = env['line'] - 1

    # Loop over each grain size
    for ia in range(Na):
        a = a_tab[ia]
        
        # Use tumbling case only for disklike grains
        if a < grainparams.a2:
            power_per_grain = dP_dnu_dOmega(env, a, beta, ip, Ndipole, tumbling=tumbling)
        else:
            power_per_grain = dP_dnu_dOmega(env, a, beta, 2/3, Ndipole, tumbling=False)

        
        nu_tab_a = power_per_grain[0, :]
        emiss_a = power_per_grain[1, :]
        
        # Find the frequencies where nu_tab is within the range of nu_tab_a
        ind = np.where((nu_tab > np.min(nu_tab_a)) & (nu_tab < np.max(nu_tab_a)))[0]
        if len(ind) > 0:
            # Interpolate emissivity and accumulate
            emiss[ind] += np.exp(interp1d(np.log(nu_tab_a), np.log(emiss_a), kind='cubic')(np.log(nu_tab[ind]))) * size_dist(a, line) * a * Da_over_a
    
    return emiss



class gff_data:
    gamma2_tab = None
    u_tab = None 
    gff_tab = None

def read_gaunt_factor():

    # Load the data using pandas
    gff_data_file = SpDust_data_dir + 'gff.dat'
    data = pd.read_csv(gff_data_file, delim_whitespace=True, comment=';', header=None, names=['gamma2', 'u', 'gff'])

    # Define dimensions
    Ngamma2 = 41
    Nu = 81

    # Initialize arrays
    gff_data.gamma2_tab = np.zeros(Ngamma2)
    gff_data.u_tab = np.zeros(Nu)
    gff_data.gff_tab = np.zeros((Ngamma2, Nu))

    # Assign values to gamma2_tab and u_tab
    gff_data.gamma2_tab = data['gamma2'].values[Nu * np.arange(Ngamma2)].astype(float)
    gff_data.u_tab = data['u'].values[0:Nu].astype(float)

    # Fill the gff_tab array
    for i in range(Ngamma2):
        gff_data.gff_tab[i, :] = data['gff'].values[i * Nu + np.arange(Nu)].astype(float)
    
    print('Gaunt factor stored')

# Call the function to test it
read_gaunt_factor()

def gaunt_factor(gamma2, u):
    # Number of elements in gamma2_tab
    Ngamma2 = len(gff_data.gamma2_tab)

    # Determine the index based on gamma2 value
    if gamma2 >= np.max(gff_data.gamma2_tab):
        index = Ngamma2 - 1
    elif gamma2 <= np.min(gff_data.gamma2_tab):
        index = 0
    else:
        # Find the largest index where gamma2_tab is less than gamma2
        index = np.max(np.where(gff_data.gamma2_tab < gamma2)[0])
        # Adjust index based on the logarithmic condition
        if np.log(gff_data.gamma2_tab[index + 1] / gamma2) < np.log(gamma2 / gff_data.gamma2_tab[index]):
            index += 1

    # Extract the corresponding row from gff_tab
    gff_new = gff_data.gff_tab[index, :]

    # Interpolate the value for the given 'u' using u_tab and gff_new
    result = np.interp(u, gff_data.u_tab, gff_new)

    return result

def free_free(env, nu_tab):
    """
    Returns j_nu/nH in cgs units (ergs/s/sr/Hatom) for free-free emission, for the given environment.
    
    Parameters:
    - env: an object or dictionary containing environment properties nh, T, xh, xC
    - nu_tab: array of frequencies in Hz
    
    Returns:
    - free-free emission rate j_nu/nH in cgs units
    """
    
    # Extract values from the environment object or dictionary
    nh = env['nh']
    T = env['T']
    xh = env['xh']
    xC = env['xC']

    me = cgsconst.me
    k = cgsconst.k
    h = cgsconst.h
    eV = cgsconst.eV
    c = cgsconst.c
    q = cgsconst.q

    # Calculate the factor
    factor = (2**5 * np.pi * q**6 / (3 * me * c**3)) * np.sqrt(2 * np.pi / (3 * k * me)) / (4 * np.pi)

    # Rydberg constant in erg (13.6 eV converted to ergs)
    Ry = 13.6 * eV

    # Gamma2 and u calculations
    gamma2 = Ry / (k * T)  # Assuming Zion = 1
    u = h * nu_tab / (k * T)

    # Call the gaunt_factor function
    gff = gaunt_factor(gamma2, u)

    # Return the free-free emission rate
    result = factor * (xh + xC)**2 * nh / np.sqrt(T) * np.exp(-h * nu_tab / (k * T)) * gff
    return result



import numpy as np
from scipy.interpolate import griddata
from scipy.interpolate import RegularGridInterpolator

class cgsconst:
    # Stores the needed constants in cgs units
    pi = 3.1415926535897932385
    c = 2.99792458e10  # speed of light
    q = 4.8032068e-10  # elementary charge
    k = 1.380650e-16   # Boltzmann constant
    mp = 1.6726231e-24  # proton mass
    me = 9.1093898e-28  # electron mass
    h = 6.6260688e-27   # Planck constant
    debye = 1e-18       # 1 Debye in cgs units
    eV = 1.60217653e-12  # 1eV in ergs

#@njit
def maketab(xmin, xmax, Npt):
    """Creates a linearly spaced table."""
    return xmin + (xmax - xmin) / Npt * (0.5 + np.arange(Npt))

#@njit
def makelogtab(xmin, xmax, Npt):
    """Creates a logarithmically spaced table."""
    return xmin * np.exp(1 / Npt * np.log(xmax / xmin) * (0.5 + np.arange(Npt)))

#@njit
def DX_over_X(xmin, xmax, Npt):
    """Returns Dx/x for a logarithmically spaced table created by makelogtab."""
    return np.exp(0.5 / Npt * np.log(xmax / xmin)) - np.exp(-0.5 / Npt * np.log(xmax / xmin))


def biinterp_func(f_val, X, Y):
    """
    Returns a function that interpolates f given at X, Y (logarithmically spaced).

    Parameters:
    - f_val: 2D array of values at Nx * Ny grid.
    - X, Y: Arrays of logarithmically spaced values.

    Returns:
    - Function that interpolates f at Xnew, Ynew.
    """


    # Interpolation requires coordinates as pairs
    return RegularGridInterpolator((X, Y), f_val, method='linear', bounds_error=False, fill_value=None)

def coord_grid(X, Y):
    """
    Returns the grid of coordinates for X, Y.
    """
    xg, yg = np.meshgrid(X, Y, indexing='ij')
    points = np.empty((xg.shape[0], xg.shape[1], 2))
    points[:, :, 0] = xg
    points[:, :, 1] = yg
    return points

def biinterplog(f, X, Y, Xnew, Ynew):
    """
    Interpolates the 2D array `f` given at logarithmically spaced points `X` and `Y`
    at new points `Xnew` and `Ynew`.
    
    Parameters:
    - f: 2D array of shape (Nx, Ny), function values at grid points.
    - X: 1D array of length Nx, original X-axis (logarithmic spacing).
    - Y: 1D array of length Ny, original Y-axis (logarithmic spacing).
    - Xnew: 1D array, new X-axis points to interpolate to.
    - Ynew: 1D array, new Y-axis points to interpolate to.
    
    Returns:
    - Interpolated values at the specified `Xnew` and `Ynew` points.
    """
    

    # Convert X and Y to logarithmic space
    logX = np.log(X)
    logY = np.log(Y)
    
    # Create indices for interpolation based on X and Y
    Xind = np.arange(len(X))
    Yind = np.arange(len(Y))
    
    # Interpolate indices for the new X and Y in log space
    Xind_new = np.interp(np.log(Xnew), logX, Xind)
    Yind_new = np.interp(np.log(Ynew), logY, Yind)
    Xind_new = np.array(Xind_new).reshape(-1,)
    Yind_new = np.array(Yind_new).reshape(-1,)
    
    # Set up 2D grid interpolator on the original index grid
    interpolator = RegularGridInterpolator((Xind, Yind), f, method='linear')
    
    # Prepare points for interpolating to `Xind_new`, `Yind_new`
    points = np.array([[xi, yi] for xi in Xind_new for yi in Yind_new])
    
    # Interpolate and reshape result
    interpolated_values = interpolator(points)
    return interpolated_values.reshape(np.size(Xnew), np.size(Ynew))

def log_biinterplog(f, X, Y, Xnew, Ynew):
    """
    Same as biinterplog, but f is logarithmically sampled.

    Parameters:
    - f: 2D array of logarithmically sampled values at Nx * Ny grid.
    - X, Y: Arrays of logarithmically spaced values.
    - Xnew, Ynew: New values at which to interpolate f.

    Returns:
    - Interpolated value of exp(f) at Xnew, Ynew.
    """

    # Perform bi-logarithmic interpolation on log(f) and exponentiate the result
    return np.exp(biinterplog(np.log(f), X, Y, Xnew, Ynew))

def log_interp(x, x_vals, y_vals):
    log_y_vals = np.log(y_vals)
    interp_log_y = np.interp(np.log(x), np.log(x_vals), log_y_vals)
    return np.exp(interp_log_y)

def extend_to_array(data):
    """ Convert list of (a, beta_tab, prob_tab) into a single array of shape (Npoints, 3). """
    extended_data = []
    for a, beta_tab, prob_tab in data:
        for beta, prob in zip(beta_tab, prob_tab):
            extended_data.append([a, beta, prob])
    return np.array(extended_data)

def readcol(name, comment=None, format=None):
    """
    Read a free-format ASCII file with columns of data into numpy arrays.

    Parameters:
        name (str): Name of ASCII data file.
        comment (str, optional): Single character specifying comment character.
                                 Any line beginning with this character will be skipped.
                                 Default is None (no comment lines).
        fmt (str, optional): Scalar string containing a letter specifying an IDL type
                              for each column of data to be read.
                              Default is None (all columns assumed floating point).

    Returns:
        The numpy arrays containing columns of data.

    Raises:
        ValueError: If invalid format string is provided.

    Note:
        This function does not support all features of the IDL `readcol` function,
        such as /SILENT, /DEBUG, /NAN, /PRESERVE_NULL, /QUICK, /SKIPLINE, /NUMLINE,
        /DELIMITER, /COUNT, /NLINES, /STRINGSKIP, or /COMPRESS keywords.
    """

    # Open the file and read lines
    with open(name, 'r') as file:
        lines = file.readlines()

    # Initialize variables
    data = [[] for _ in range(len(lines[0].split()))]  # List to store data columns

    fmt = format.split(', ') if format else None

    # Process each line
    for line in lines:
        # Skip comment lines
        if comment and line.strip().startswith(comment):
            continue

        # Split the line into fields based on whitespace
        fields = line.strip().split()[:len(fmt)]
        
        # Convert each field based on format
        for i, field in enumerate(fields):
            if fmt:
                fmt_type = fmt[i].upper()
            else:
                fmt_type = 'F'  # Default to floating point if format not specified

            if fmt_type == 'A':
                data[i].append(field)
            elif fmt_type in ('D', 'F', 'I', 'B', 'L', 'U', 'Z'):
                data[i].append(float(field))
            elif fmt_type == 'X':
                continue  # Skip this column
            else:
                raise ValueError(f"Invalid format type: {fmt_type}")

    # Convert lists to numpy arrays
    arrays = [np.array(column) for column in data if len(column) > 0]

    return arrays

# Example usage:
# v1, v2, v3 = readcol("data.txt", fmt="F,F,A")





from .Grain import N_C, N_H, grain_distribution
from .util import cgsconst, makelogtab
from .mpiutil import *
import numpy as np

from .SED import mu2_f, SED

debye = cgsconst.debye

def SpyDust(environment, tumbling=True, output_file=None, min_freq=None, max_freq=None, n_freq=None, Ndipole=None, single_beta=False, spdust_plasma=False):

    # Check the environment structure for required parameters
    if 'dipole' not in environment and 'dipole_per_atom' not in environment:
        print("Please specify the dipole moment (of a=1e-7 cm) or dipole_per_atom.")
        return
    
    # Determine the dipole moment
    if 'dipole' in environment:
        mu_1d_7 = environment['dipole']
        dipole_per_atom = mu_1d_7 / np.sqrt(N_C(1e-7) + N_H(1e-7)) * debye
    else:
        dipole_per_atom = environment['dipole_per_atom'] * debye
        mu_1d_7 = np.sqrt(N_C(1e-7) + N_H(1e-7)) * environment['dipole_per_atom']
        

    # Check for grain size distribution parameters
    if 'line' not in environment:
        print("Please specify the grain size distribution parameters (Weingartner & Draine, 2001a).")
        return

    line = environment['line']-1

    # Number of dipole moments
    Ndip = 20
    if Ndipole is not None:
        Ndip = Ndipole

    # Set in-plane moment ratio
    ip = 2 / 3
    if 'inplane' in environment:
        print(f"Assuming that <mu_ip^2>/<mu^2> = {environment['inplane']} for disklike grains")
        ip = environment['inplane']

    # Frequency settings
    GHz = 1e9
    numin = 1 * GHz
    numax = 100 * GHz
    Nnu = 200
    if min_freq is not None:
        numin = min_freq * GHz
    if max_freq is not None:
        numax = max_freq * GHz
    if n_freq is not None:
        Nnu = n_freq

    nu_tab = makelogtab(numin, numax, Nnu)

    grain_obj = grain_distribution()
    f_a_beta = grain_obj.shape_and_size_dist(line, normalize=False, fixed_thickness=single_beta)
    a_tab = grain_obj.a_tab
    beta_tab = grain_obj.beta_tab

    ang_Omega_min = 1e7
    ang_Omega_max = 1e15
    N_angular_Omega = 1000
    angular_Omega_tab = makelogtab(ang_Omega_min, ang_Omega_max, N_angular_Omega)

    mu2_f_arr = mu2_f(environment, a_tab, beta_tab, f_a_beta, 
                      dipole_per_atom, ip, Ndip, 
                      tumbling=tumbling, 
                      parallel=True, 
                      contract_a =True, 
                      omega_min=ang_Omega_min, 
                      omega_max=ang_Omega_max, 
                      Nomega=N_angular_Omega,
                      spdust_plasma=spdust_plasma)
    
    cos_theta_list = np.linspace(-1, 1, 20)

    # We mimic spdust and give a quite ad-hoc treatment to the angular distribution of the grain rotation.
    # The user can provide a more sophisticated treatment of the distribution of internal alignment.
    cos_theta_weights = []
    for beta in beta_tab:
        if beta > 0: # rotation around the axis of largest inertia (prolate grain, theta = pi/2)
            aux = np.zeros_like(cos_theta_list)
            aux[15] = 1
            cos_theta_weights.append(aux)
        elif beta > -0.1: # rotation around the axis of largest inertia (nearly spherical grains; theta=0)
            aux = np.zeros_like(cos_theta_list)
            aux[-1] = 1
            cos_theta_weights.append(aux)
        elif tumbling:
            cos_theta_weights.append(None) # rotation with isotropic distribution of theta
        else:  # rotation around the axis of largest inertia (oblate grains; theta=0)
            aux = np.ones_like(cos_theta_list)
            aux[-1] = 1
            cos_theta_weights.append(aux)

    resultSED = SED(nu_tab, mu2_f_arr, beta_tab, angular_Omega_tab, cos_theta_list, cos_theta_weights)

     # Handle free-free emission if requested
    Jy = 1e-23
    result = np.zeros((2, Nnu))

    result[0, :] = nu_tab / GHz
    result[1, :] = resultSED / Jy

    # Write output to file
    if rank0:
        if output_file is not None:        
            with open(output_file, 'w') as f:
                f.write('#=========================== SPYDUST ===============================\n')
                f.write(f'#    nH = {environment["nh"]} cm^-3\n')
                f.write(f'#    T = {environment["T"]} K\n')
                f.write(f'#    Chi = {environment["Chi"]}\n')
                f.write(f'#    xh = {environment["xh"]}\n')
                f.write(f'#    xC = {environment["xC"]}\n')
                f.write(f'#    mu(1E-7 cm) = {mu_1d_7} debye (beta = {dipole_per_atom / debye} debye)\n')
                if tumbling:
                    f.write('#    Disklike grains are randomly oriented with respect to angular momentum.\n')
                else:
                    f.write('#    Disklike grains spin around their axis of greatest inertia\n')
                f.write('#=====================================================================\n')

                np.savetxt(f, result.T, fmt='%12.6e') # Columns are nu, jnu_per_H, jnu_per_H_freefree
    
    return result # shape (2, Nnu); rows are nu, SED(nu) in Jy

    
    

def SpyDust_single_grain(environment, a, beta, tumbling=True, min_freq=None, max_freq=None, n_freq=None, Ndipole=None):

    # Check the environment structure for required parameters
    if 'dipole' not in environment and 'dipole_per_atom' not in environment:
        print("Please specify the dipole moment (of a=1e-7 cm) or dipole_per_atom.")
        return
    
    # Determine the dipole moment
    if 'dipole' in environment:
        mu_1d_7 = environment['dipole']
        dipole_per_atom = mu_1d_7 / np.sqrt(N_C(1e-7) + N_H(1e-7)) * debye
    else:
        dipole_per_atom = environment['dipole_per_atom'] * debye
        mu_1d_7 = np.sqrt(N_C(1e-7) + N_H(1e-7)) * environment['dipole_per_atom']
        

    # Check for grain size distribution parameters
    if 'line' not in environment:
        print("Please specify the grain size distribution parameters (Weingartner & Draine, 2001a).")
        return

    line = environment['line']-1

    # Number of dipole moments
    Ndip = 20
    if Ndipole is not None:
        Ndip = Ndipole

    # Set in-plane moment ratio
    ip = 2 / 3
    if 'inplane' in environment:
        print(f"Assuming that <mu_ip^2>/<mu^2> = {environment['inplane']} for disklike grains")
        ip = environment['inplane']

    # Frequency settings
    GHz = 1e9
    numin = 1 * GHz
    numax = 100 * GHz
    Nnu = 200
    if min_freq is not None:
        numin = min_freq * GHz
    if max_freq is not None:
        numax = max_freq * GHz
    if n_freq is not None:
        Nnu = n_freq

    nu_tab = makelogtab(numin, numax, Nnu)

    f_a_beta = np.array([1]).reshape(1, 1)
    a_tab = np.array([a])
    beta_tab = np.array([beta])

    ang_Omega_min = 1e7
    ang_Omega_max = 1e15
    N_angular_Omega = 1000
    angular_Omega_tab = makelogtab(ang_Omega_min, ang_Omega_max, N_angular_Omega)

    mu2_f_arr = mu2_f(environment, a_tab, beta_tab, f_a_beta, 
                      dipole_per_atom, ip, Ndip, 
                      tumbling=tumbling, 
                      parallel=True, 
                      contract_a =True, 
                      omega_min=ang_Omega_min, 
                      omega_max=ang_Omega_max, 
                      Nomega=N_angular_Omega,
                      spdust_plasma=False)
    
    cos_theta_list = np.linspace(-1, 1, 20)

    # We mimic spdust and give a quite ad-hoc treatment to the angular distribution of the grain rotation.
    # The user can provide a more sophisticated treatment of the distribution of internal alignment.
    cos_theta_weights = []
    for beta in beta_tab:
        if beta > 0: # rotation around the axis of largest inertia (prolate grain, theta = pi/2)
            aux = np.zeros_like(cos_theta_list)
            aux[15] = 1
            cos_theta_weights.append(aux)
        elif beta > -0.1: # rotation around the axis of largest inertia (nearly spherical grains; theta=0)
            aux = np.zeros_like(cos_theta_list)
            aux[-1] = 1
            cos_theta_weights.append(aux)
        elif tumbling:
            cos_theta_weights.append(None) # rotation with isotropic distribution of theta
        else:  # rotation around the axis of largest inertia (oblate grains; theta=0)
            aux = np.ones_like(cos_theta_list)
            aux[-1] = 1
            cos_theta_weights.append(aux)

    resultSED = SED(nu_tab, mu2_f_arr, beta_tab, angular_Omega_tab, cos_theta_list, cos_theta_weights)

     # Handle free-free emission if requested
    Jy = 1e-23
    result = np.zeros((2, Nnu))

    result[0, :] = nu_tab / GHz
    result[1, :] = resultSED / Jy
    
    return result # shape (2, Nnu); rows are nu, SED(nu) in Jy

   


    



# Initialisation file for the 'SpyDust' package.

# Initialize package variables
package_version = "1.1"

# Importing submodules for ease of access
import os
SpDust_data_dir = os.path.join(__path__[0], 'Data_Files/')
#SpDust_data_dir = os.path.join("/home/phil/software/SpyDust/SpyDust", 'Data_Files/')


from . import AngMomDist, charge_dist, collisions, free_free, Grain, \
              H2_photoemission, infrared, mpiutil, plasmadrag, SED, \
              util



from . import SpDust_data_dir
from .util import cgsconst

import numpy as np
import pandas as pd

pi = np.pi
c = cgsconst.c
h = cgsconst.h
k = cgsconst.k
eV = cgsconst.eV
me = cgsconst.me
q = cgsconst.q
mp = cgsconst.mp


class gff_data:
    gamma2_tab = None
    u_tab = None 
    gff_tab = None

def read_gaunt_factor():

    # Load the data using pandas
    gff_data_file = SpDust_data_dir + 'gff.dat'
    data = pd.read_csv(gff_data_file, delim_whitespace=True, comment=';', header=None, names=['gamma2', 'u', 'gff'])

    # Define dimensions
    Ngamma2 = 41
    Nu = 81

    # Initialize arrays
    gff_data.gamma2_tab = np.zeros(Ngamma2)
    gff_data.u_tab = np.zeros(Nu)
    gff_data.gff_tab = np.zeros((Ngamma2, Nu))

    # Assign values to gamma2_tab and u_tab
    gff_data.gamma2_tab = data['gamma2'].values[Nu * np.arange(Ngamma2)].astype(float)
    gff_data.u_tab = data['u'].values[0:Nu].astype(float)

    # Fill the gff_tab array
    for i in range(Ngamma2):
        gff_data.gff_tab[i, :] = data['gff'].values[i * Nu + np.arange(Nu)].astype(float)
    
    print('Gaunt factor stored')

# Call the function to test it
read_gaunt_factor()

def gaunt_factor(gamma2, u):
    # Number of elements in gamma2_tab
    Ngamma2 = len(gff_data.gamma2_tab)

    # Determine the index based on gamma2 value
    if gamma2 >= np.max(gff_data.gamma2_tab):
        index = Ngamma2 - 1
    elif gamma2 <= np.min(gff_data.gamma2_tab):
        index = 0
    else:
        # Find the largest index where gamma2_tab is less than gamma2
        index = np.max(np.where(gff_data.gamma2_tab < gamma2)[0])
        # Adjust index based on the logarithmic condition
        if np.log(gff_data.gamma2_tab[index + 1] / gamma2) < np.log(gamma2 / gff_data.gamma2_tab[index]):
            index += 1

    # Extract the corresponding row from gff_tab
    gff_new = gff_data.gff_tab[index, :]

    # Interpolate the value for the given 'u' using u_tab and gff_new
    result = np.interp(u, gff_data.u_tab, gff_new)

    return result

def free_free(env, nu_tab):
    """
    Returns j_nu/nH in cgs units (ergs/s/sr/Hatom) for free-free emission, for the given environment.
    
    Parameters:
    - env: an object or dictionary containing environment properties nh, T, xh, xC
    - nu_tab: array of frequencies in Hz
    
    Returns:
    - free-free emission rate j_nu/nH in cgs units
    """
    
    # Extract values from the environment object or dictionary
    nh = env['nh']
    T = env['T']
    xh = env['xh']
    xC = env['xC']


    # Calculate the factor
    factor = (2**5 * np.pi * q**6 / (3 * me * c**3)) * np.sqrt(2 * np.pi / (3 * k * me)) / (4 * np.pi)

    # Rydberg constant in erg (13.6 eV converted to ergs)
    Ry = 13.6 * eV

    # Gamma2 and u calculations
    gamma2 = Ry / (k * T)  # Assuming Zion = 1
    u = h * nu_tab / (k * T)

    # Call the gaunt_factor function
    gff = gaunt_factor(gamma2, u)

    # Return the free-free emission rate
    result = factor * (xh + xC)**2 * nh / np.sqrt(T) * np.exp(-h * nu_tab / (k * T)) * gff
    return result



from .. import SpDust_data_dir
from ..util import cgsconst, makelogtab, DX_over_X
from .grain_properties import grainparams, N_C, acx
from .charge_dist import Qabs, nu_uisrf
from .infrared import Temp, Energy_modes, IR_arrays

import numpy as np
from scipy.special import erf
import os


h = cgsconst.h
c = cgsconst.c
q = cgsconst.q
k = cgsconst.k
eV = cgsconst.eV
mp = cgsconst.mp

pi = np.pi


# Define the Tev function
def Tev(a, Chi):
    """
    Calculate the equilibrium temperature for a given grain size `a` and radiation field `Chi`.
    
    Parameters:
    - a: Grain size.
    - Chi: Radiation field strength.

    Returns:
    - Temperature (Tev) in K.
    """

    # Thermal spikes limit
    hnu_tab_vals = makelogtab(1e-2, 13.6, 500)  
    Qabs_vals = Qabs(a, 1, hnu_tab_vals)        
    nu_uisrf_vals = nu_uisrf(hnu_tab_vals)      

    E_photon = (np.sum(Qabs_vals * nu_uisrf_vals) / 
                np.sum(Qabs_vals * nu_uisrf_vals / hnu_tab_vals)) * eV

    Energy_modes_op, Energy_modes_ip, Energy_modes_ch  = Energy_modes(a)     
    T_q = Temp(a, np.array([E_photon]), Energy_modes_op, Energy_modes_ip, Energy_modes_ch)  

    # Steady emission limit (Take cross-section of ionized PAH)
    Dnu_over_nu = DX_over_X(1e-2, 13.6, 500)  
    Qu_star = Chi * np.sum(Qabs_vals * nu_uisrf_vals) * Dnu_over_nu

    hnu_0 = np.array([1e-4]) 
    Qlambda_0 = Qabs(a, 1, hnu_0) * (c / (hnu_0 * eV / h))**2

    # Constants for zeta(6) and gamma(6)
    zeta6 = np.pi**6 / 945
    gamma6 = 120

    # Calculate steady state temperature T_c
    T_c = h * c / k * (Qu_star / (8 * np.pi * h * c * Qlambda_0 * gamma6 * zeta6))**(1 / 6)

    # Return the maximum of T_q and T_c
    return max(T_q, T_c)

    
a_min, a_max, Na, chi_min, chi_max, Nchi \
= IR_arrays.a_min, IR_arrays.a_max, IR_arrays.Na, IR_arrays.chi_min, IR_arrays.chi_max, IR_arrays.Nchi
FIR_integral_charged, GIR_integral_charged, FIR_integral_neutral,GIR_integral_neutral \
= IR_arrays.FIR_integral_charged, IR_arrays.GIR_integral_charged, IR_arrays.FIR_integral_neutral, IR_arrays.GIR_integral_neutral

a_tab = IR_arrays.a_tab
chi_tab = IR_arrays.chi_tab

def compute_Tev():
    """
    Compute the Tev array and save it to a file.
    
    Parameters:
    - SpDust_data_dir: Directory path where the output file will be saved.
    - a_min, a_max: Minimum and maximum grain sizes.
    - Na: Number of grain size intervals.
    - chi_min, chi_max: Minimum and maximum Chi values.
    - Nchi: Number of Chi intervals.
    """

    output_file = os.path.join(SpDust_data_dir, "Tev_{}a_{}chi.txt".format(Na, Nchi))

    # If the Tev table already stored, load it and return
    if os.path.exists(output_file):
        print(f"Tev table already exists at {output_file}.")
        IR_arrays.Tev_tab = np.loadtxt(output_file)
        return
    
    # Initialize the Tev_tab array (Na x Nchi)
    Tev_tab = np.zeros((Na, Nchi))

    # Nested loops to compute Tev for each combination of `a` and `Chi`
    for ia in range(Na):
        a = a_tab[ia]
        for ichi in range(Nchi):
            Chi = chi_tab[ichi]
            Tev_tab[ia, ichi] = Tev(a, Chi)  

    # Save the result to a file
    # output_file = f"{SpDust_data_dir}/Tev_{Na}a_{Nchi}chi.txt"
    
    
    # Save using numpy's savetxt function
    np.savetxt(output_file, Tev_tab, fmt='%.6e')
    IR_arrays.Tev_tab = Tev_tab
    print(f"Tev table saved to {output_file}")
    return 

# Call the compute_Tev function
compute_Tev()

def Tev_interpol(a, Chi):
    """
    Perform interpolation to compute Tev for given grain size `a` and radiation field `Chi`.

    Parameters:
    - a: Grain size.
    - Chi: Radiation field strength.

    Returns:
    - Interpolated Tev value.
    """

    # Finding the indices and coefficients for `a`
    if a <= np.min(a_tab):
        ia = 0
        alpha = 1.0
    elif a >= np.max(a_tab):
        ia = Na - 2
        alpha = 0.0
    else:
        ia = np.max(np.where(a_tab <= a)[0])
        alpha = 1.0 - Na * np.log(a / a_tab[ia]) / np.log(a_max / a_min)

    # Finding the indices and coefficients for `Chi`
    if Chi <= np.min(chi_tab):
        print(f'Warning: you are using Chi={Chi}! The code is written for Chi > 1E-6. If you wish to extend its validity to lower values, set chi_min to a lower value, run "compute_Tev", and recompile.')
        ichi = 0
        beta = 1.0
    elif Chi >= np.max(chi_tab):
        print(f'Warning: you are using Chi={Chi}! The code is written for Chi < 1E8. If you wish to extend its validity to higher values, set chi_max to a higher value, run "compute_Tev", and recompile.')
        ichi = Nchi - 2
        beta = 0.0
    else:
        ichi = np.max(np.where(chi_tab <= Chi)[0])
        beta = 1.0 - Nchi * np.log(Chi / chi_tab[ichi]) / np.log(chi_max / chi_min)

    # Interpolating the Tev value
    Tev_log_interp = (
        alpha * (beta * np.log(IR_arrays.Tev_tab[ia, ichi]) + (1.0 - beta) * np.log(IR_arrays.Tev_tab[ia, ichi + 1])) +
        (1.0 - alpha) * (beta * np.log(IR_arrays.Tev_tab[ia + 1, ichi]) + (1.0 - beta) * np.log(IR_arrays.Tev_tab[ia + 1, ichi + 1]))
    )

    # Return the exponentiated result
    return np.exp(Tev_log_interp)

# Define the Tev_effective function
def Tev_effective(env, a, get_info=False):
    """
    Compute the effective evaporation temperature.

    Parameters:
    - env: A dictionary containing environment parameters like nh, T, Chi, Tev (optional).
    - a: Grain size.
    - get_info: If True, return the ratio R_{coll/abs}/Nsites instead of the temperature.

    Returns:
    - Effective evaporation temperature (Tev) or ratio if get_info is True.
    """

    a2, d = grainparams.a2, grainparams.d
    
    # Check if the constant evaporation temperature is present
    if 'Tev' in env:
        return env['Tev']

    # Extract environmental parameters
    nh = env['nh']
    T = env['T']
    Chi = env['Chi']

    # Generate the necessary arrays
    hnu_tab_vals = makelogtab(1e-2, 13.6, 500)  
    Dnu_over_nu = DX_over_X(1e-2, 13.6, 500) 
    Qabs_vals = Qabs(a, 1, hnu_tab_vals)  
    nu_uisrf_vals = nu_uisrf(hnu_tab_vals)  

    # Calculate the ratio of collision rate to absorption rate
    ratio = nh * np.sqrt(8 * k * T / (pi * mp)) / (
        Chi * np.sum(Qabs_vals * nu_uisrf_vals / (hnu_tab_vals * eV)) * Dnu_over_nu * c
    )

    # Number of sites (Nsites)
    if a < a2:
        Nsites = N_C(a)  
    else:
        Nsites = N_C(a) * 3 * d / a

    # If `get_info` is True, return the ratio/Nsites
    if get_info:
        return ratio / Nsites

    # Determine the effective temperature
    if Nsites > ratio:
        Tev = Tev_interpol(a, Chi)  
    else:
        Tev = T  # No more sticking collisions, atoms bounce back

    return Tev

# Define the FGn function
def FGn(env, a, T_ev, Zg_tab, tumbling=True):
    """
    Calculate F_n and G_n for neutral impactors, for an array of grain charges.
    
    Parameters:
    - env: Dictionary containing environment parameters like T, xh, y.
    - a: Grain size.
    - T_ev: Precomputed evaporation temperature.
    - Zg_tab: Array of grain charges.
    - tumbling: Boolean flag for tumbling condition.

    Returns:
    - Dictionary with Fn and Gn values.
    """

    #--- polarizabilities for neutral species : Hydrogen, Helium and H2 ---

    a0       = 0.52918e-8      # Bohr radius
    alpha_H  = 4.5 * a0**3      # J. Chem. Phys. 49, 4845 (1968); DOI:10.1063/1.1669968  
    alpha_He = 1.38 * a0**3     # M A Thomas et al 1972 J. Phys. B: At. Mol. Phys. 5 L229-L232 
    alpha_H2 = 5.315 * a0**3    # W.C. Marlow, Proc. Phys. Soc 1965, vol. 86 (ground state)
    
    # Extract environmental parameters
    Tval = env['T']
    xh = env['xh']
    y = env['y']
    acx_val = acx(a)  

    # Calculate e_n and e_e
    e_n = np.sqrt(q**2 / (2 * acx_val**4 * k * Tval) * Zg_tab**2)
    e_e = np.sqrt(q**2 / (2 * acx_val**4 * k * T_ev) * Zg_tab**2)

    # Contribution of atomic Hydrogen
    eps_n = e_n * np.sqrt(alpha_H)
    eps_e = e_e * np.sqrt(alpha_H)
    FnH_ev = (1 - xh - y) * (np.exp(-eps_n**2) + np.sqrt(pi) * eps_n * erf(eps_n)) / \
             (np.exp(-eps_e**2) + np.sqrt(pi) * eps_e * erf(eps_e)) * \
             (np.exp(-eps_e**2) + 2 * eps_e**2)
    GnH_in = 0.5 * (1 - xh - y) * (np.exp(-eps_n**2) + 2 * eps_n**2)

    if tumbling:
        FnH_in = FnH_ev * 2 / 3 / (1 + np.sqrt(2 / 3) * eps_n)
    else:
        FnH_in = 0.0

    # Contribution of neutral Helium
    eps_n = e_n * np.sqrt(alpha_He)
    eps_e = e_e * np.sqrt(alpha_He)
    FnHe_ev = 1 / 6 * (np.exp(-eps_n**2) + np.sqrt(pi) * eps_n * erf(eps_n)) / \
              (np.exp(-eps_e**2) + np.sqrt(pi) * eps_e * erf(eps_e)) * \
              (np.exp(-eps_e**2) + 2 * eps_e**2)
    GnHe_in = 1 / 12 * (np.exp(-eps_n**2) + 2 * eps_n**2)

    if tumbling:
        FnHe_in = FnHe_ev * 2 / 3 / (1 + np.sqrt(2 / 3) * eps_n)
    else:
        FnHe_in = 0.0

    # Contribution of molecular Hydrogen
    eps_n = e_n * np.sqrt(alpha_H2)
    eps_e = e_e * np.sqrt(alpha_H2)
    FnH2_ev = y / np.sqrt(2) * (np.exp(-eps_n**2) + np.sqrt(np.pi) * eps_n * erf(eps_n)) / \
              (np.exp(-eps_e**2) + np.sqrt(np.pi) * eps_e * erf(eps_e)) * \
              (np.exp(-eps_e**2) + 2 * eps_e**2)
    GnH2_in = 0.5 * y / np.sqrt(2) * (np.exp(-eps_n**2) + 2 * eps_n**2)

    if tumbling:
        FnH2_in = FnH2_ev * 2 / 3 / (1 + np.sqrt(2 / 3) * eps_n)
    else:
        FnH2_in = 0.0

    # Sum of contributions
    Fn_ev = FnH_ev + FnHe_ev + FnH2_ev
    Fn_in = FnH_in + FnHe_in + FnH2_in
    Fn = Fn_ev + Fn_in
    Gn = GnH_in + GnHe_in + GnH2_in + 0.5 * T_ev / Tval * Fn_ev

    return {'Fn': Fn, 'Gn': Gn}

def FGn_averaged(env, a, T_ev, fZ, tumbling=True):
    """
    Compute the averaged Fn and Gn over grain charges given the grain charge distribution.
    
    Parameters:
    - env: Dictionary containing environment parameters like T, xh, y.
    - a: Grain size.
    - T_ev: Precomputed evaporation temperature.
    - fZ: 2D array with grain charges in fZ[0, *] and charge distribution in fZ[1, *].
    - tumbling: Boolean flag for tumbling condition.

    Returns:
    - Dictionary with averaged Fn and Gn values.
    """
    
    # Call FGn to get Fn and Gn for specific grain charges
    FGn_result = FGn(env, a, T_ev, fZ[0, :], tumbling=tumbling)

    # Calculate averaged Fn and Gn
    Fn_averaged = np.sum(FGn_result['Fn'] * fZ[1, :])
    Gn_averaged = np.sum(FGn_result['Gn'] * fZ[1, :])

    return {'Fn': Fn_averaged, 'Gn': Gn_averaged}

def g1g2(psi, mu_tilde):
    """
    Compute g1 and g2 based on the given psi and mu_tilde array.
    
    Parameters:
    - psi: Scalar value representing psi.
    - mu_tilde: Array of values representing mu_tilde (size Ndipole).

    Returns:
    - Dictionary with g1 and g2 arrays.
    """

    # Number of elements in mu_tilde
    Ndipole = np.size(mu_tilde)

    # if type is float, convert to array
    if isinstance(mu_tilde, (int, float)):
        mu_tilde = np.array([mu_tilde])

    # Initialize g1 and g2 arrays
    g1 = np.zeros(Ndipole)
    g2 = np.zeros(Ndipole)

    # Index where mu_tilde <= abs(psi)
    index = np.where(mu_tilde <= abs(psi))[0]
    if len(index) > 0:
        mu_i = mu_tilde[index]
        if psi < 0:
            g1[index] = 1 - psi
            g2[index] = 1 - psi + 0.5 * psi**2 + (1 / 6) * mu_i**2
        else:
            if psi > 600:  # To avoid floating underflow
                g1[index] = 0
                g2[index] = 0
            else:
                g1[index] = np.exp(-psi) * np.sinh(mu_i) / mu_i
                g2[index] = g1[index]

    # Index where mu_tilde > abs(psi)
    index = np.where(mu_tilde > abs(psi))[0]
    if len(index) > 0:
        mu_i = mu_tilde[index]
        g1[index] = (1 - np.exp(-(psi + mu_i)) + mu_i - psi + 0.5 * (mu_i - psi)**2) / (2 * mu_i)
        g2[index] = g1[index] + (mu_i - psi)**3 / (12 * mu_i)

    # Return a dictionary with g1 and g2
    return g1, g2

# Define the h1 function
def h1(phi, mu_tilde):
    """
    Compute h1(phi, mu_tilde) based on AHD09 Eq.(103).
    
    Parameters:
    - phi: Scalar value representing phi.
    - mu_tilde: Array or scalar representing mu_tilde.

    Returns:
    - h1: Computed value based on the formula. Shape depends on the input.
    """

    mu = mu_tilde

    # Calculate u_0
    u_0 = 0.5 * (-phi + np.sqrt(phi**2 + 4 * mu))

    # Compute the coefficients
    DL98 = 1 + np.sqrt(pi) / 2 * phi
    coeff_1 = mu / 4 + phi**2 / (4 * mu) + (1 - mu) / (2 * mu)
    coeff_erf = np.sqrt(pi) * phi * (3 - 2 * mu) / (8 * mu)
    coeff_exp = -(4 + phi**2 + phi * np.sqrt(phi**2 + 4 * mu)) / (8 * mu)

    # Compute the result for h1
    h1_result = DL98 + coeff_1 + coeff_erf * erf(u_0) + coeff_exp * np.exp(-u_0**2)

    return h1_result

def h2(phi, mu_tilde):
    """
    Compute h2(phi, mu_tilde) as per AHD09 Eq.(104).
    
    Parameters:
    - phi: Scalar value representing phi.
    - mu_tilde: Array of values representing mu_tilde.

    Returns:
    - Array of h2 values corresponding to the input phi and mu_tilde.
    """
    
    mu = mu_tilde

    # Calculate intermediate values
    u_0 = 0.5 * (-phi + np.sqrt(phi**2 + 4 * mu))
    Dl98 = 1 + 3 * np.sqrt(np.pi) / 4 * phi + 0.5 * phi**2
    coeff_1 = mu**2 / 12 + mu / 4 + phi**2 / (2 * mu) + (1 - mu) / (2 * mu) - phi**2 / 4
    coeff_erf = np.sqrt(np.pi) * phi / (32 * mu) * (4 * mu**2 - 12 * mu + 15 + 2 * phi**2)
    coeff_exp = (phi**2 * (2 * mu - 9) - 16 + (2 * mu - 7) * phi * np.sqrt(phi**2 + 4 * mu)) / (32 * mu)

    # Compute h2
    h2_values = Dl98 + coeff_1 + coeff_erf * erf(u_0) + coeff_exp * np.exp(-u_0**2)

    return h2_values

# Define the FGi function
def FGi(env, a, T_ev, Zg, mu_tab):
    """
    Compute Fi and Gi for a given grain charge Zg, based on mu_tab.
    
    Parameters:
    - env: Dictionary containing environment parameters like T, xh, xC (assumed to represent xM in the original code).
    - a: Grain size.
    - T_ev: Precomputed evaporation temperature.
    - Zg: Grain charge.
    - mu_tab: Array of mu values.

    Returns:
    - Dictionary with Fi and Gi values.
        Fi: shape (len(mu_tab),) array
    """

    # Extract environmental parameters
    T = env['T']
    xh = env['xh']
    xM = env['xC']  

    # Compute acx and mu_tilde
    acx_val = acx(a)  # Assuming acx(a) is defined elsewhere
    mu_tilde = q * mu_tab / (acx_val**2 * k * T)

    # Polarizabilities for neutral species corresponding to incoming ions (H and C)
    a0 = 0.53e-8  # Bohr radius in cm
    alpha_H = 4.5 * a0**3  # Polarizability for Hydrogen
    alpha_C = 1.54e-24  # Polarizability for Carbon

    # Neutral grains case
    if Zg == 0:
        phi = np.sqrt(2) * q / np.sqrt(acx_val * k * T)
        Fi = (xh + xM * np.sqrt(12)) * h1(phi, mu_tilde)  
        Gi_in = 0.5 * (xh + xM * np.sqrt(12)) * h2(phi, mu_tilde)  
    else:
        # Charged grains case
        psi = Zg * q**2 / (acx_val * k * T)
        e_i = np.sqrt(q**2 / (2 * acx_val**4 * k * T_ev) * Zg**2)

        # Contribution of H+ and C+
        eps_i = e_i * np.sqrt([alpha_H, alpha_C])
        Fi_contrib = [xh, xM] * (np.exp(-eps_i**2) + 2 * eps_i**2) / (np.exp(-eps_i**2) + np.sqrt(np.pi) * eps_i * erf(eps_i))
        g1, g2 = g1g2(psi, mu_tilde)  
        Fi = np.sum(Fi_contrib) * g1
        Gi_in = 0.5 * (xh + xM * np.sqrt(12)) * g2

    # Calculate Gi_ev and return Fi and Gi
    Gi_ev = 0.5 * T_ev / T * Fi
    Gi = Gi_in + Gi_ev

    return {'Fi': Fi, 'Gi': Gi}

def FGi_averaged(env, a, T_ev, mu_tab, fZ):
    """
    Compute the averaged Fi and Gi over grain charges given the grain charge distribution.
    
    Parameters:
    - env: Dictionary containing environment parameters like T, xh, xC (representing xM).
    - a: Grain size.
    - T_ev: Precomputed evaporation temperature.
    - mu_tab: Array of mu values.
    - fZ: 2D array where fZ[0, *] contains grain charges and fZ[1, *] contains their respective probabilities.

    Returns:
    - Dictionary with averaged Fi and Gi arrays.
    """

    # Initialize Fi and Gi as arrays of zeros with the same dimensions as mu_tab
    Fi = np.zeros_like(mu_tab).reshape(-1)
    Gi = np.zeros_like(mu_tab).reshape(-1)


    # Number of grain charge values in fZ
    NZg = fZ.shape[1]

    # Loop over all grain charges in fZ[0, *]
    for i in range(NZg):
        # Compute Fi and Gi for the current grain charge fZ[0, i]
        FGi_result = FGi(env, a, T_ev, fZ[0, i], mu_tab)

        # Average Fi and Gi by summing over the charge distribution
        Fi += fZ[1, i] * FGi_result['Fi']
        Gi += fZ[1, i] * FGi_result['Gi']

    # Return the averaged Fi and Gi
    return {'Fi': Fi, 'Gi': Gi}



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name="SpyDust",  # Required
    version="1.1",  # Required
    description="A code for modeling spinning dust radiation",  # Optional
    long_description=long_description,  # Optional
    long_description_content_type="text/markdown",  # Optional (see note above)
    url="https://github.com/zzhang0123/SpyDust",  # Optional
    author="Zheng Zhang",  # Optional
    author_email="zheng.zhang@manchester.ac.uk",  # Optional
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[  # Optional
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    keywords="astrophysics, astronomy",  # Optional
    #package_dir={"": "SpyDust"},  # Optional
    packages=find_packages(include=["SpyDust", 
                                    "SpyDust.SPDUST_as_is",
                                    #"SpyDust.core", 
                                    #"SpyDust.main", 
                                    #"SpyDust.utils"
                                    ]),  # Required
    python_requires=">=3.7, <4",
    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/discussions/install-requires-vs-requirements/
    install_requires=[
        "numpy", 
        "scipy",
        "logging",
        "mpi4py",
        #"pandas" # not necessary for spinning dust spectra; 
                  # but if free-free emission is also desired (using free_free.py), then it is needed
        ],
    extras_require={},
    zip_safe=False, # don't compile the package into a zip file when installed
    # If there are data files included in your packages that need to be
    # installed, specify them here.
    package_data={'SpyDust': ['Data_Files/*']},
)

